//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

//#include "Riostream.h"
//#include <TApplication.h>
//#include <TString.h>
#include <TROOT.h>
#include <TFile.h>
#include <TCanvas.h>
#include <TGraph.h>
#include <TLegend.h>
#include <TMultiGraph.h>
#include <TH1F.h>
#include <TH2F.h>
#include <TPad.h>
#include <TView.h>
#include <TStyle.h>
#include <TAxis.h>

#include <QString>
#include <QSettings>
#include "bpmlib/bpmappsettings.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <deque>
#include <vector>
#include <map>
#include <list>
#include <cmath>
#include <locale>
#include <cassert>
#include <algorithm>

#include <time.h>
#include <wchar.h>

#include "bpmdata.h"
#include "bpmlib/bpmconst.h"
#include "util.h"

#include <unistd.h> // daemon()

using std::string;
using std::wstring;
using std::cout;
using std::wcout;
using std::cerr;
using std::wcerr;
using std::endl;
using std::list;
using std::map;
using std::vector;
using std::deque;

string fileBase = "orbita";
string fmt = "png";

const QString appName = "BPMGUI";
const QString settingsRoot = QString("/")+appName;
BpmAppSettings settings("AFI Electronics", appName);

int runAnalysis(const string &inputFileName)
{
    const BpmRun &bpmRun = BpmData::getReadOrbitaCsv(inputFileName);

    if (bpmRun.size() < 1) {
        cerr << "No input data" << endl;
	return 1;
    }

    const size_t eventCount = bpmRun.size();
    if (eventCount < 1) {
        cerr << "No events found" << endl;
        return 1;
    } else {
        cout << eventCount << " events found" << endl;
    }

    bool backgroundEventFound = false;
    int backgroundEvent = 0;
    // select signal/background by power
    double minLevel = signalLevelNoiseThreshold;
    double maxLevel = signalLevelNoiseThreshold;
    const BpmSignalLevels &bpmLevels = BpmData::getEventSignalLevels(bpmRun);
    if (!bpmLevels.empty()) {
        const BpmSignalLevels::const_iterator min_el = std::min_element(bpmLevels.begin(), bpmLevels.end(), BpmSignalLevel_value_comp);
        const BpmSignalLevels::const_iterator max_el = std::max_element(bpmLevels.begin(), bpmLevels.end(), BpmSignalLevel_value_comp);
        minLevel = min_el->second;
        maxLevel = max_el->second;
        const int minEvId = min_el->first;
        const int maxEvId = max_el->first;
        cout << "min signal level: " << minLevel  << " dB at event " << minEvId << endl;
        cout << "max signal level: " << maxLevel  << " dB at event " << maxEvId  << endl;
        if (maxLevel - minLevel > minSignalToBackgroundRatioDb) {
            backgroundEventFound = true;
            backgroundEvent = minEvId;
            cout << "Background event " << minEvId << endl;
        } else {
            cout << "Background event not found" << endl;
        }
        //    for(BpmRun::const_iterator it = bpmRun.begin(); it != bpmRun.end(); ++it) {
        //        const int ev = it->first;
        //        const BpmEvent &curEvent = it->second;
        //    }
    }
    map<int, bool> enableEvents;
    list<int> suppressedEvents, passedEvents;
    int numEventsSuppressed = 0;
    int numEventsTotal = 0;
    for(BpmSignalLevels::const_iterator it = bpmLevels.begin(); it!=bpmLevels.end(); ++it) {
        int ev = it->first;
        double level = it->second;
        bool enable = true;
        numEventsTotal ++;
        if (backgroundEventFound) {
            if (level < minLevel + signalLevelOverBackgroundThresholdDb) {
                enable = false;
                numEventsSuppressed++;
                //                cout << "Suppressed event " << ev << endl;
            } else {
                //                cout << "Passed event " << ev << endl;
            }
        }
        enableEvents[ev] = enable;
        if (enable) passedEvents.push_back(ev); else suppressedEvents.push_back(ev);
    }
    if (suppressedEvents.empty()) {
        cout << "Passed all events" << endl;
    } else {
        cout << "Suppressed events: ";
        for(list<int>::const_iterator it = suppressedEvents.begin(); it != suppressedEvents.end(); ++it)
            cout << " " << *it;
        cout << endl;
        cout << "Passed events: ";
        for(list<int>::const_iterator it = passedEvents.begin(); it != passedEvents.end(); ++it)
            cout << " " << *it;
        cout << endl;
    }

    const BpmEvent &firstEvent = bpmRun.begin()->second;
    const size_t deviceCount = firstEvent.size();
    cout << deviceCount << " devices in first event" << endl;
    //    for(BpmEvent::const_iterator it = firstEvent.begin(); it != firstEvent.end(); ++it) {
    //        wcerr << (*it).first << L": " << (*it).second.label << endl;
    //    }
    for(BpmRun::const_iterator it = bpmRun.begin(); it != bpmRun.end(); ++it) {
        const int ev = it->first;
        const BpmEvent &curEvent = it->second;
        if (curEvent.size() != deviceCount) {
            cerr << "Device count mismatch in event " << ev << ": " << curEvent.size() << endl;
            for(BpmEvent::const_iterator it = curEvent.begin(); it != curEvent.end(); ++it)
                wcerr << (*it).first << L": " << (*it).second.label << endl;
            enableEvents[ev] = false;
            //            return 1;
        }
    }

    const string timeTag = timeStr(getFileMtime(inputFileName));
    TGraph *graphs[4][eventCount];
    for(int j=0; j<4; j++)
        for(unsigned int ev=0; ev<eventCount; ev++)
            graphs[j][ev] = 0;
    Double_t x[4][eventCount][deviceCount+1], y[4][eventCount][deviceCount+1];
    unsigned int ev = 0;
    for(BpmRun::const_iterator it = bpmRun.begin(); it != bpmRun.end(); ++it) {
        assert(ev < eventCount);
        const int eventId = it->first;
        if (!enableEvents[eventId]) {
            //            cout << "Suppressed event " << eventId << endl;
            //            continue;
        }
        const BpmEvent &curEvent = it->second;
        typedef std::map<double, wstring> MapType;
        MapType m;
        for(BpmEvent::const_iterator it = curEvent.begin(); it != curEvent.end(); ++it) {
            const BpmEntry &bpmEntry = it->second;
            const wstring &label = bpmEntry.label;
            double pickupPos = bpmEntry.pos;
            m[pickupPos] = label;
        }
        int posCount[4] = {0, 0, 0, 0};
        double BField = 0;

        for(MapType::const_iterator it = m.begin(); it != m.end(); ++it) {
            //            const double pickupPos = it->first;
            const wstring &label = it->second;
            const BpmEntry &bpmEntry = curEvent.at(label);
            BField = bpmEntry.field;

            // mute channels without signal or with wrong signal
            bool skipR = false;
            bool skipZ = false;
            if (bpmEntry.powerR < signalLevelNoiseThreshold) {
                //                wcerr << "Event " << ev << ": no R signal on " <<  bpmEntry.label << ": " << bpmEntry.powerR << " dB" << endl;
                skipR = true;
            }
            if (bpmEntry.powerZ < signalLevelNoiseThreshold) {
                //                wcerr << "Event " << ev << ": no Z signal on " <<  bpmEntry.label << ": " << bpmEntry.powerZ << " dB" << endl;
                skipZ = true;
            }
            const double level = bpmLevels.at(bpmEntry.event);
            if (fabs(bpmEntry.powerR - level) > powerDbMaxMismatch) {
                skipR = true;
            }
            if (fabs(bpmEntry.powerZ - level) > powerDbMaxMismatch) {
                skipZ = true;
            }
            //            if (fabs(bpmEntry.powerZ - bpmEntry.powerR) > powerDbMaxMismatch) {
            //                wcerr << "Event " << ev << ": R/Z signal mismatch on " <<  bpmEntry.label << ": " << bpmEntry.powerR << ", " << bpmEntry.powerZ << " dB" << endl;
            //                skipR = true;
            //                skipZ = true;
            //            }

            // set data
            for(int j=0; j<4; j++) {
                bool db = (j%2)==1;
                const int pc = posCount[j];
                if (x_is_metres) {
                    x[j][ev][pc] = bpmEntry.pos + positionShift;
                } else {
                    x[j][ev][pc] = 1+pc;
                }
                if(j==0)
                    y[j][ev][pc] = bpmEntry.posR * pickupSize[0];
                if(j==1)
                    y[j][ev][pc] = bpmEntry.powerR;
                if(j==2)
                    y[j][ev][pc] = bpmEntry.posZ * pickupSize[1];
                if(j==3)
                    y[j][ev][pc] = bpmEntry.powerZ;
            }
            // increment plot point position
            if (!skipR) {
                posCount[0]++;
            }
            posCount[1]++;
            if (!skipZ) {
                posCount[2]++;
            }
            posCount[3]++;
        }

        for(int j=0; j<4; j++) {
            const int pc = posCount[j];
            if (pc > 1) {
                // mirror first measurement to last+1
                if (x_is_metres) {
                    x[j][ev][pc] = orbitLength + positionShift;
                } else {
                    x[j][ev][pc] = x[j][ev][pc] + 1;
                }
                y[j][ev][pc] = y[j][ev][0];
                posCount[j]++;
            }
            // create Graphs
            if (pc > 1) {
                graphs[j][ev] = new TGraph(pc+1, x[j][ev], y[j][ev]);
                graphs[j][ev]->SetTitle(TString::Format("%.1f Gs", BField));
//                graphs[j][ev]->SetTitle(TString::Format("%.4f T", BField*1e-4));
            } else {
                graphs[j][ev] = 0;
            }
        }
        ev++;
    }

    std::ostringstream ostr;
    if (backgroundEventFound) {
        ostr << "S/N: " << 0.1*int(10.*maxLevel) << "/" << 0.1*int(10.*minLevel) << " dB" << " (" << numEventsTotal-numEventsSuppressed << "/" << numEventsTotal << ")";
    } else {
        ostr << 0.1*int(10.*maxLevel) << " dB";
    }
    string levelTag = ostr.str();
    const TString title = TString("Orbit   ") + timeTag + "   " + levelTag;
    const TString titleX = orbitLength?"Pickup position, m":"Pickup number";

    string imageFileName[4];
    string imageFileNameTmp[4];
    for(int h2d = 0; h2d < 2; h2d++)
        for(int rz = 0; rz < 4; rz++) {
        bool db = (rz%2)==1;
        TString titleY;
        if (db)
            titleY = "Beam " + TString((rz/2)?"Z":"R") + " Intensity, dB";
        else
            titleY = "Beam " + TString((rz/2)?"Z":"R") + " Position, mm";
        TCanvas *c = new TCanvas("c1", "c1", (int)(1.5*480), 380);
        c->SetGrid();
        c->SetLeftMargin(0.08);
        c->SetRightMargin(0.04);
//        TMultiGraph *mg = new TMultiGraph();
        TH1F *h = 0;
        TH2F *h2 = 0;

        double x1, x2, y1, y2;
        if (x_is_metres) {
            x1 = 0; x2 = orbitLength;
        } else {
            x1 = 1; x2 = deviceCount;
        }
        if (db) {
            y1 = signalLevelAbsoluteMinimumDb; y2 = 0;
        } else {
            y1 = -pickupSize[rz/2]*1.1; y2 = pickupSize[rz/2]*1.1;
        }

        if (0==h2d) {
            h = c->DrawFrame(x1, y1, x2, y2);
            for(unsigned int ev=0; ev<eventCount; ev++) {
                if (!enableEvents[ev] && !db)
                    continue;
                TGraph *g = graphs[rz][ev];
                if (!g) continue;
//                    mg->Add(g, "LP");
                    double eventLevel = bpmLevels.at(ev);
                    if (eventLevel >= signalLevelStrongThreshold)
                        g->SetLineStyle(1);
                    else
                        if (eventLevel >= signalLevelWeakThreshold)
                            g->SetLineStyle(2);
                    else
                        g->SetLineStyle(3);

                    g->SetMarkerStyle(21);
                    g->SetMarkerSize(0.5);
                    g->SetMarkerColor(kBlue);
                    g->SetLineColor(kBlack);
                    g->SetFillStyle(0);
                    g->Draw("LP");
                    h->GetXaxis()->SetTitle(titleX);
                    h->GetXaxis()->SetNdivisions(8, kFALSE);
                    h->GetYaxis()->SetTitle(titleY);
                    h->SetTitle(title);
            }
//            mg->Draw();
//            mg->SetTitle(title);
//            assert(mg->GetXaxis());
//            assert(mg->GetYaxis());
//            mg->GetXaxis()->SetLimits(x1, x2);
//            mg->GetXaxis()->SetTitle(titleX);
//            mg->GetXaxis()->SetNdivisions(8, kFALSE);
//            mg->GetYaxis()->SetLimits(y1, y2); // does not work
//            mg->GetYaxis()->SetTitle(titleY);

        } else {
            //            int nbinsx = deviceCount;
            //            double xlow = 0;
            //            double xup = deviceCount;
            //            int nbinsy = eventCount;
            //            double ylow = 0;
            //            double yup = eventCount;
            //            h2 = new TH2F("h2", "Title", nbinsx, xlow, xup, nbinsy, ylow, yup);
            //            for(int ev=0; ev<eventCount; ev++)
            //                for(int dev=0; dev<deviceCount; dev++) {
            //                double v = y[rz][ev][dev];
            //                if (db) {
            //                    v = y[rz][ev][dev];
            //                    if (v < dbOffset) v = dbOffset;
            //                    v -= dbOffset;
            //                } else {
            //                }
            //                h2->SetBinContent(1+dev, 1+ev, v);
            //            }
            //            h2->Draw("LEGO2");
        }

        // create legend if possible
        double lh = 0.05*eventCount;
        if (lh>0.4) lh = 0.4;
        TLegend *l = c->BuildLegend(0.82, 0.9-lh, 0.96, 0.9);
        if (l) {
            l->Clear();
            l->SetFillColor(kWhite);
            for(unsigned int ev=0; ev<eventCount; ev++) {
                if (!enableEvents[ev] && !db)
                    continue;
                TGraph *g = graphs[rz][ev];
                if (!g) continue;
                l->AddEntry(g, "", "LP");
            }
        }

        // print
        TVirtualPad *p = c->Pad();
        assert(p);
        assert(p->GetFrame());
        p->Update();
        const char *rzTags[4] = { "r", "rdb", "z", "zdb" };
        string rzTag = rzTags[rz];
        if (h2d) rzTag += string("2");

        imageFileName[rz] = fileBase + "-" + rzTag + string(".")+fmt;
        imageFileNameTmp[rz] = fileBase + "-" + rzTag + string(".tmp.")+fmt;
        p->Print(imageFileNameTmp[rz].c_str(), fmt.c_str());
        delete h;
        delete h2;
        delete c;

        unlink(imageFileName[rz].c_str());
        rename(imageFileNameTmp[rz].c_str(), imageFileName[rz].c_str());
    }

    return 0;
}

void orbitaweb()
{
    runAnalysis(fileBase+".csv");
}

int runAnalysisLoop(const char *inputFileName)
{
    time_t last_mtime = 0;
    while(true) {
        time_t mtime = getFileMtime(inputFileName);
        if (mtime != last_mtime) {
            last_mtime = mtime;
            cout << endl;
            runAnalysis(inputFileName);
        }
        usleep(fileStatPeriodUs);
    }
    return 0;
}

void rootSucks()
{
    gROOT->SetStyle("Plain");
    gStyle->SetPalette (1, 0);
    gStyle->SetOptStat(0);
    gStyle->SetTitleAlign(23); // holy shit
    gStyle->SetTitleX(0.5);
    gStyle->SetTitleBorderSize(0);
    gStyle->SetTitleFontSize(0.06);
}

int main(int argc, char *argv[])
{
    std::locale::global(std::locale(""));
    if (argc < 2) {
        cerr << "USAGE: " << argv[0] << " [-d] <filename.csv> [gif|png|svg]" << endl;
        return 1;
    }
    bool daemonize = false;
    std::string filename;
    for(int i=1; i<argc; i++) {
        string argstr = argv[i];
        if (argstr == "gif" || argstr == "png" || argstr == "svg") {
            fmt = argstr;
//            cerr << "Set image format " << fmt << endl;
        } else if (argstr == "-d") {
            daemonize = true;
//            cerr << "Run as daemon" << endl;
        } else {
            filename = argstr;
//            cerr << "filename: " << filename << endl;
        }
    }
    if (daemonize) daemon(1, 0);

    settings.read();
    if (0 != runAnalysis(filename)) {
        cerr << "Failed to start" << endl;
        return 1;
    }
    rootSucks();

    runAnalysisLoop(argv[1]);
    return 0;
}
