//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "bpmdata.h"

#include <iostream>
#include <fstream>
#include <vector>
#include <deque>
#include <sstream>
#include <cmath>
#include <algorithm>

#include "util.h"
#include "bpmlib/bpmconst.h"

using std::string;
using std::wifstream;
//using std::string;
using std::wstring;
using std::cout;
//using std::wcout;
using std::cerr;
using std::wcerr;
using std::endl;
//using std::map;
using std::vector;
using std::deque;

BpmData::BpmData(const string &inputFileName)
    : fileName(inputFileName)
{
}

/// read orbita.csv file
/// line format: device,title,pos,event,field,posH,powerH,posV,powerV
BpmRun BpmData::getReadOrbitaCsv(const string &inputFileName)
{
    BpmRun bpmRun;
    std::wifstream ifs(inputFileName.c_str());
    if (!ifs.good()) {
        cerr << "Can't open file " << inputFileName << endl;
        return BpmRun();
    }

    {
        std::vector<wstring> lineData;
        std::wstring str;
        int lineNumber = 0;
        while (ifs.good() && !ifs.eof() && std::getline(ifs, str, L'\n')) {
            lineNumber++;
            if (lineNumber == 1) {
                lineNumber++;
                continue;
            }
            lineData.clear();
            std::wistringstream in(str);
            //        in.sync();
            BpmEntry bpmEntry;
            wstring sub;
            while(std::getline(in, sub, L',')) {
                lineData.push_back(sub);
            }

            if (lineData.size() < 9) {
                std::wcerr << lineNumber << ": insufficient number of values: " << str << endl;
                break;
            }
            const wstring label = lineData[0];
            bpmEntry.label  = label;
            bpmEntry.place  = lineData[1];
            bpmEntry.pos  = wstring_to_double(lineData[2]);
            int event = wstring_to_int(lineData[3]);
            bpmEntry.event  = event;
            bpmEntry.field  = wstring_to_double(lineData[4]);
            bpmEntry.posR   = wstring_to_double(lineData[5]);
            bpmEntry.powerR = wstring_to_double(lineData[6]);
            bpmEntry.posZ   = wstring_to_double(lineData[7]);
            bpmEntry.powerZ = wstring_to_double(lineData[8]);
            if (lineData.size() >= 10) {
                bpmEntry.bgrR = wstring_to_double(lineData[9]);
                bpmEntry.bgrZ = wstring_to_double(lineData[10]);
            }
            bpmRun[event][label] = bpmEntry;
            if (lineNumber % 1000 == 0) {
                cout << "Read " <<  lineNumber << " lines" << endl;
            }
        }
        cout << "Read " <<  lineNumber << " lines" << endl;
    }
    return bpmRun;
}

bool BpmSignalLevel_value_comp(const std::pair<int, double>& lhs,
                               const std::pair<int, double>& rhs)
{
    return lhs.second < rhs.second;
}

double BpmSignalLevels::min() const
{
    if (empty()) return signalLevelAbsoluteMinimumDb;
    return std::min_element(begin(), end(), BpmSignalLevel_value_comp)->second;
}

double BpmSignalLevels::max() const
{
    if (empty()) return signalLevelAbsoluteMinimumDb;
    return std::max_element(begin(), end(), BpmSignalLevel_value_comp)->second;
}

BpmSignalLevels BpmData::getEventSignalLevels(const BpmRun &bpmRun)
{
    BpmSignalLevels result;
    for(BpmRun::const_iterator it = bpmRun.begin(); it != bpmRun.end(); ++it) {
        const int ev = it->first;
        const BpmEvent &curEvent = it->second;
        deque<double> levels;
        for(BpmEvent::const_iterator it = curEvent.begin(); it != curEvent.end(); ++it) {
            const BpmEntry &bpmEntry = it->second;
//            const wstring &label = bpmEntry.label;
            double powerR = bpmEntry.powerR;
            double powerZ = bpmEntry.powerZ;
//            if (powerR < signalLevelNoiseThreshold) continue;
//            if (powerZ < signalLevelNoiseThreshold) continue;
            if (fabs(powerR - powerZ) > powerDbMaxMismatch) continue;
            double level = sumDb(powerR, powerZ) - 6.02;
            levels.push_back(level);
        }
        if (levels.size() > 6) {
            levels.pop_front();
            levels.pop_front();
            levels.pop_front();
            levels.pop_back();
            levels.pop_back();
            levels.pop_back();
        }
        double power = 0;
        if (levels.size() > 0) {
            for(size_t i=0; i<levels.size(); i++)
                power += fromDb(levels[i]);
            power /= levels.size();
            power = toDb(power);
        }
        result[ev] = power;
    }
    return result;
}
