//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QSettings>
#include <QStringList>
#include <QDebug>

#include "bpmconst.h"
#include "bpmappsettings.h"
#include "beamsetup.h"

start_select_t StartSelect::fromVariant(const QVariant &v)
{
    return fromString(v.toString());
}

start_select_t StartSelect::fromString(const QString &s)
{
    start_select_t t = BPMStartSoftware; // default
    if (s == "B0") t = BPMStartB0;
    if (s == "Software") t = BPMStartSoftware;
    if (s == "Ext") t = BPMStartExt;
    return t;
}

QString StartSelect::toString(const start_select_t &t)
{
    QString s;
    switch(t) {
    case BPMStartB0:
        s = "B0";
        break;
    case BPMStartExt:
        s = "Ext";
        break;
    case BPMStartSoftware:
        // fall through
    default:
        s = "Software";
    }
    return s;
}


trigger_select_t TriggerSelect::fromVariant(const QVariant &v)
{
    return fromString(v.toString());
}

trigger_select_t TriggerSelect::fromString(const QString &s)
{
    trigger_select_t t = BPMTriggerPeriodic; // default
    if (s == "TableBField") t = BPMTriggerTableBField;
    if (s == "TableTime") t = BPMTriggerTableTime;
    if (s == "Periodic") t = BPMTriggerPeriodic;
    if (s == "Ext") t = BPMTriggerExt;
    return t;
}

QString TriggerSelect::toString(const trigger_select_t &t)
{
    QString s;
    switch(t) {
    case BPMTriggerTableBField:
        s = "TableBField";
        break;
    case BPMTriggerTableTime:
        s = "TableTime";
        break;
    case BPMTriggerExt:
        s = "Ext";
        break;
    case BPMTriggerPeriodic:
        // fall through
    default:
        s = "Periodic";
    }
    return s;
}

BpmAppSettings::BpmAppSettings( const QString & organization, const QString & application )
{
    _organization = organization;
    _application = application;
    setDefaults();
}

// translate list of QStrings to list of QVariants
QList<QVariant> BpmAppSettings::translateStringListToVariantList(const QStringList &sl) {
    QList<QVariant> vlist;
    foreach(QString str, sl) {
        vlist.push_back(str);
    }
    return vlist;
}

// translate list of QVariants to list of QStrings (when possible)
QStringList BpmAppSettings::translateVariantListToStringList(const QVariant &var)
{
    QStringList sl;
    if (var.type() != QVariant::List) {
        qDebug() << QString("Not a list: ") << var;
        return sl;
    }
    QList<QVariant> vlist(var.toList());
    foreach(QVariant v, vlist) {
        sl += v.toString();
    }
    return sl;
}

void BpmAppSettings::read(const QString &fileName)
{
    if (fileName.isNull()) {
        QSettings qsettings(_organization, _application);
        const QStringList &keys = qsettings.allKeys();
        foreach(QString key, keys) {
            (*this)[key] = qsettings.value(key);
        }
    } else {
        QSettings qsettings(fileName, QSettings::IniFormat);
        const QStringList &keys = qsettings.allKeys();
        foreach(QString key, keys) {
            (*this)[key] = qsettings.value(key);
        }
    }
    //  settings.print("readSettings()");
    //    updateBeamSetup();
}

void BpmAppSettings::write(const QString &fileName)
{
    if (fileName.isNull()) {
        QSettings qsettings(_organization, _application);
        foreach(QString key, keys()) {
            qsettings.setValue(key, (*this).value(key));
        }
    } else {
        QSettings qsettings(fileName, QSettings::IniFormat);
        foreach(QString key, keys()) {
            qsettings.setValue(key, (*this).value(key));
        }
    }
}

void BpmAppSettings::setDefaults()
{
    (*this)["EnableH"] = true;
    (*this)["EnableV"] = true;
    (*this)["EnableS"] = true;
    (*this)["EnableD"] = true;
    (*this)["SelectEvent"] = 0;
    (*this)["EnableRun"] = true;
    (*this)["EnableWriteRaw"] = false;
    (*this)["MeasureFrequency"] = false;
    (*this)["SpillWait"] = 0;
    (*this)["SpillLength"] = 5.0;
    (*this)["SampleCount"] = 65536;
    (*this)["MaxSampleCount"] = MaxSampleCount;
    (*this)["TriggerSelect"] = "TableBField";
    (*this)["SumFFT"] = false;

    // Pickup geometry
    (*this)["PickupHSize"] = pickupSize[0];
    (*this)["PickupVSize"] = pickupSize[1];

    // Nuclotron
    {
        AppSettings m;
        m["kf"] = 5;
        m["rho"] = 22.0;
        m["Circumference"] = orbitLength;
        setWithPrefix("Nuclotron", m);
    }

    // Beam
    {
        AppSettings m;
        m["W0"] = d_W0;
        setWithPrefix("Beam", m);
    }
    //    updateBeamSetup();

    // B values are stored in Teslas
    (*this)["BMin"] = -20000.0 * 1e-4;
    (*this)["BMax"] = 20000.0 * 1e-4;
    (*this)["B0"] = 100.0 * 1e-4;
    (*this)["BStep"] = 0.1 * 1e-4;
    (*this)["TriggerTableBField"] = "{ 295.0, 300.0, 1000.0 }";
    (*this)["TriggerTableTime"] = "{ 0.001, 0.1, 0.5 }";

    (*this)["periodicTrigDelay"] = 0.0;
    (*this)["periodicTrigPeriod"] = 0.1;
}

QList<quint32> BpmAppSettings::getSampleSizeList() const
{
    QList<quint32> list;
    // hardware may be not ready at this moment, take hardcoded default
    quint32 max_samples = value("MaxSampleCount").toUInt();
    quint32 radix=0;
    while((1u<<(radix+1)) < max_samples) radix++;
    //  size_t s = 1<<radix;

    list.push_back(max_samples);

    for(quint32 r=radix; r>3; r--) {
        quint32 s = 1<<r;
        list.push_back(s);
    }
    return list;
}
