//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <cmath>
//#include <iostream>
//#include <sstream>
//#include <iomanip>
//#include <cerrno>

#include <QDebug>
#include <QFile>
#include <QDataStream>
#include <QVariant>
#include <QSettings>
#include <QStringList>

#include "dsplib/dsplib.h"
#include "dsplib/fftengine.h"
#include "dsplib/VectorMaths.h"
#include "misclib/TextFormatter.h"

//#include "bpmbox.h"
#include "bpmevent.h"
#include "bpmboxproperties.h"
#include "misclib/appsettings.h"
#include "devicelabels.h"


//void BPMBoxEvent::print() const
//{
//    std::cout << (*this) << std::endl;
//}

//void DaqEvent::print() const
//{
//    std::cout << (*this) << std::endl;
//}

QTextStream& operator<<(QTextStream &s, const DaqEventBoxAdc &daqEventBoxAdc)
{
    double rms = daqEventBoxAdc.rms;
    s << "Mean" << daqEventBoxAdc.baseline
            << ", RMS " << rms
            << "   " << val2db(rms*1.414/8192) << " dBFS";
    //    s << daqEventBoxAdc.getRmsText();
    //  samples.print();
    return s;
}

QTextStream& operator<<(QTextStream &s, const DaqEventBox &daqEventBox)
{
    s << "DaqEventBox: ID 0x"
            << hex << daqEventBox.id
            << dec << ", clockFreqFactor "
            << daqEventBox.clockFreqFactor
            << ", temp " << daqEventBox.temp
            << endl;
//    s << daqEventBox.adc << endl;
    return s;
}

QTextStream& operator<<(QTextStream &s, const DaqEvent &ev)
{
    s << "DaqEvent " << ev.eventNumber << ": " << ev.box.size() << " stations";
//    s << ev.box;
    return s;
}

QTextStream& operator<<(QTextStream &s, const BPMBoxEvent &ev)
{
    s << "BPMBoxEvent at " << ev.timeStamp << " s, " << ev.BField_Gs << " Gs, " << ev.adc.size() << " raw samples";
    return s;
}

QString DaqEventBox::getFullName() const
{
    int id = getId();
    return QString("BPM %1 '%2'").arg(id).arg(deviceLabel(id));
}

double DaqEvent::getMasterTimeStamp() const
{
    if (!box.contains(masterId))
        return 0;
    return box[masterId].getTimeStamp();
}

DaqEventBox::DaqEventBox()
{
    id = 0;
    clockFreqFactor = defaultClockFreqFactor;
    temp = 0;
    timeStamp = 0;
}

//void DaqEventBox::print() const
//{
//    std::cout << (*this) << std::endl;
//}
//void DaqEventBoxAdc::print() const
//{
//    std::cout << (*this) << std::endl;
//}

void DaqEventBoxAdc::setData(const DaqAdcSamples &r)
{
    samples = r;
    updateStats();
    for(DaqAdcSamples::iterator it = samples.begin(); it!=samples.end(); ++it) {
        (*it) -= baseline;
    }
    //  samples = samples - baseline;
}

void DaqEventBox::setData(const DaqEventBoxAdcs &r)
{
    adc = r;
}

void DaqEventBoxAdc::resample()
{
    double k = getClockFreqFactor();
    if (k != 0 && fabs(k) < 0.01) {
        qDebug() << "Resampling by 1 - " << k;
        samples = DaqAdcSamples::fromStdVector(RFFTEngine::resample(samples.toStdVector(), 1-k, 1, 1)); // linear, no oversample
    }
}
/*
DaqAdcSamples& DaqAdcSamples::operator=(const myvector<double>&r)
{
  *this = DaqAdcSamples(r);
  return *this;
}
*/

void DaqEventBox::setClockFreqFactor(double v)
{
    clockFreqFactor = v;
    for(DaqEventBoxAdcs::iterator it=adc.begin(); it!=adc.end(); ++it) {
        (*it).setClockFreqFactor(v);
    }
}

void DaqEventBoxAdc::updateStats()
{
    updateRms();
}


void DaqEventBoxAdc::updateRms()
{
    // Subtract baseline - TODO
    baseline = VectorMaths::calculateMean(samples.toStdVector());
    rms = VectorMaths::calculateRMS(samples.toStdVector());
    //  baseline = samples.mean();
    //  rms = samples.rms();
}

//QString DaqEventBoxAdc::getRmsText() const
//{
//    std::ostringstream ost;
//    ost << "Mean" << dec << setw(8) << baseline << ", RMS " << setw(8) << rms << "   " << val2db(rms*1.414/8192) << " dBFS";
//    return ost.str();
//}

//void DaqEventBoxAdc::printRms() const
//{
//    cerr << getRmsText() << std::endl;
//}

double DaqEventBox::getSamplingFreq() const
{
    return BpmBoxProperties::sampleFreq();
}

const DaqEventBoxAdc &DaqEvent::operator[](size_t n) const
{
    Q_ASSERT(box.find(n/4) != box.end());
    const DaqEventBox &daqEventBox = box[n/4];
    Q_ASSERT(n%4 < daqEventBox.size());
    return daqEventBox[n%4];
}


/*
QString DaqEvents::getRawFileNameTimestamped(const time_t &t)
{
    char timestr[100];
    time_t timep = t;
    if (timep == 0) {
        timep = time(NULL);
    }
    strftime(timestr, sizeof(timestr), "%Y-%m-%d_%H-%M-%S_%Z", localtime (&timep));
    timestr[sizeof(timestr)-1] = 0;
    return QString("/data/")+"bpm_"+timestr+".dat";
}
*/

QDataStream &operator<<(QDataStream &stream, const DaqEventBoxAdc &daqEventBoxAdc)
{
    QMap<QString, QVariant> parameters;
    parameters["baseline"] = daqEventBoxAdc.getBaseline();
    parameters["rms"] = daqEventBoxAdc.getRms();
    parameters["clockFreqFactor"] = daqEventBoxAdc.getClockFreqFactor();
    stream << parameters;
    stream << daqEventBoxAdc.getSamples();
    return stream;
}

QDataStream &operator>>(QDataStream &stream, DaqEventBoxAdc &daqEventBoxAdc)
{
    QMap<QString, QVariant> parameters;
    stream >> parameters;
    DaqAdcSamples data;
    stream >> data;
    daqEventBoxAdc = DaqEventBoxAdc();
    daqEventBoxAdc.setClockFreqFactor(parameters["clockFreqFactor"].toDouble());
    daqEventBoxAdc.setData(data);
    return stream;
}

QDataStream &operator<<(QDataStream &stream, const DaqEventBox &daqEventBox)
{
    QMap<QString, QVariant> parameters;
    parameters["id"] = daqEventBox.getId();
    parameters["clockFreqFactor"] = daqEventBox.getClockFreqFactor();
    parameters["temp"] = daqEventBox.getTemp();
    parameters["samplingFreq"] = daqEventBox.getSamplingFreq();
    parameters["samplingFreqCorrected"] = daqEventBox.getSamplingFreqCorrected();
    parameters["timeStamp"] = daqEventBox.getTimeStamp();
    parameters["timeStampCorrected"] = daqEventBox.getTimeStampCorrected();

    stream << parameters;
    stream << daqEventBox.getAdc();
    return stream;
}

QDataStream &operator>>(QDataStream &stream, DaqEventBox &daqEventBox)
{
    QMap<QString, QVariant> parameters;
    stream >> parameters;
    DaqEventBoxAdcs data;
    stream >> data;
    daqEventBox = DaqEventBox();
    daqEventBox.setClockFreqFactor(parameters["clockFreqFactor"].toDouble());
    daqEventBox.setId(parameters["id"].toInt());
    daqEventBox.setTemp(parameters["temp"].toDouble());
    daqEventBox.setTimeStamp(parameters["timeStamp"].toDouble());
    daqEventBox.setData(data);
    return stream;
}

const char *fileMagicBPM = "BPM RAW ";
const char *fileMagicEV  = " EV ";
const char *fileMagicBOX = "BOX ";
const char *fileMagicADC = "ADC ";
const char *fileMagicEOF = "ZZZZ";

QDataStream &operator<<(QDataStream &stream, const DaqEvent &daqEvent)
{
    QMap<QString, QVariant> parameters;
    parameters["masterId"] = daqEvent.masterId;
    parameters["eventNumber"] = daqEvent.eventNumber;
    parameters["BField_Gs"] = daqEvent.BField_Gs;
    stream << parameters;
    stream << daqEvent.box;
    return stream;
}

QDataStream &operator>>(QDataStream &stream, DaqEvent &daqEvent)
{
    QMap<QString, QVariant> parameters;
    stream >> parameters;
    DaqEventBoxes data;
    stream >> data;
    daqEvent = DaqEvent();
    daqEvent.masterId = parameters["masterId"].toInt();
    daqEvent.eventNumber = parameters["eventNumber"].toInt();
    daqEvent.BField_Gs = parameters["BField_Gs"].toDouble();
    daqEvent.box = data;
    return stream;
}

QString DaqEvents::writeEvents(const AppSettings &settings, const DaqEvents &daqEvents, const QString &fileName)
{
    QFile f(fileName);
    if (!f.open(QFile::WriteOnly)) {
        qCritical() << "Can't open for writing file " << fileName << ": " << f.errorString();
        return QString();
    }
    QDataStream ds(&f);
    ds << settings;
    ds << daqEvents;
/*
    ds << fileMagicBPM;
    const quint32 fileVersion = 0x0200;
    ds << fileVersion;
    const quint32 eventCount = daqEvents.size();
    ds << eventCount;
    Q_FOREACH(DaqEvent event, daqEvents) {
        ds << fileMagicEV;
        // begin write event data
        quint32 eventNumber = event.eventNumber;
        ds << eventNumber;
        const quint32 masterId = event.masterId;
        ds << masterId;
        double eventMasterTimeStamp = event.getMasterTimeStamp();
        ds << eventMasterTimeStamp;
        double eventBField_Gs = event.BField_Gs;
        ds << eventBField_Gs;
        const DaqEventBoxes &boxes = event.box;
        const quint32 boxCount = boxes.size();
        ds << boxCount;
        Q_FOREACH(DaqEventBox box, boxes) {
            // begin write box data
            ds << fileMagicBOX;
            const quint32 id = box.getId();
            ds << id;
            double eventTimeStamp = box.getTimeStamp();
            ds << eventTimeStamp;
            double clockFreqFactor = box.getClockFreqFactor();
            ds << clockFreqFactor;
            double temp = box.getTemp();
            ds << temp;

            const quint32 adcCount = box.size();
            ds << adcCount;
            for(size_t c=0; c<box.size(); c++) {
                const DaqEventBoxAdc &adc = box[c];
                ds << fileMagicADC;
                const DaqAdcSamples &gotSamples = adc.getSamples();
                const quint32 sampleCount = gotSamples.size();
                ds << sampleCount;
                ds << gotSamples;
            }
            // end write box data
        }
        // end write event data
    }
    ds << fileMagicEOF;
*/
    if (ds.status() != QDataStream::Ok)
        qCritical() << "Error writing file '" << fileName << "': " << f.errorString();
    f.close();
    return fileName;
}
/*
bool findMagic(FILE *f, const char *magic, const QString &fileName = QString())
{
    off_t startOff = ftell(f);
    while(true) {
        if (feof(f)) {
            cerr << "End of file '" << fileName << "' while searching for '" << magic << "': " << strerror(errno) << endl;
            return false;
        }
        off_t foundOff = ftell(f);
        char readMagic[strlen(magic)+1];
        size_t nr = fread(readMagic, sizeof(readMagic), 1, f);
        if (nr != 1) {
            cerr << "Can't read file '" << fileName << "': " << strerror(errno) << endl;
            return false;
        }
//        cerr << "Test '" << readMagic << "' '" << magic << "'" << endl;
        if (strcmp(readMagic, magic) != 0)
            continue;
        if (foundOff != startOff) {
            cerr << "Skipped " << foundOff-startOff << " bytes in file '" << fileName << "' while searching for '"
                    << magic << "'" << endl;
        }
        return true;
    }
}
*/

bool DaqEvents::readRawData(const QString &fileName, DaqEvents *daqEvents, AppSettings *appSettings)
{
    Q_ASSERT(daqEvents);
    Q_ASSERT(appSettings);
    bool ok = true;
//    DaqEvents daqEvents;
    QFile f(fileName);
    if (!f.open(QFile::ReadOnly)) {
        qCritical() << "Can't open for reading file " << fileName << ": " << f.errorString();
        return false;
    }
    QDataStream ds(&f);
//    AppSettings appSettings;
    ds >> *appSettings;
    ds >> *daqEvents;
    if (ds.status() != QDataStream::Ok) {
        ok = false;
        qCritical() << "Error reading file '" << fileName << "': " << f.errorString();
    }
    f.close();
    return ok;
}

DaqEvents DaqEvents::readEvents(const QString &fileName)
{
    DaqEvents daqEvents;
    AppSettings appSettings;
    readRawData(fileName, &daqEvents, &appSettings);

/*
    quint32 fileVersion = 0;
    quint32 eventCount = 0;
    FILE *f = fopen(fileName.c_str(), "r");
    if (!f) {
        cerr << "Can't open for reading file '" << fileName << "': " << strerror(errno) << endl;
        return daqEvents;
    }
    if (!findMagic(f, fileMagicBPM, fileName)) {
        cerr << "Can't read file '" << fileName << "': " << strerror(errno) << endl;
        fclose(f);
        return daqEvents;
    }
    if (1 != fread(&fileVersion, sizeof(fileVersion), 1, f)) {
        cerr << "Can't read file '" << fileName << "': " << strerror(errno) << endl;
        goto end;
    }

    if (fileVersion != 0x0100) {
        cerr << "Unknown version " << fileVersion << ", file '" << fileName << "'" << endl;
        goto end;
    }

    if (1 != fread(&eventCount, sizeof(eventCount), 1, f)) {
        cerr << "Can't read file '" << fileName << "': " << strerror(errno) << endl;
        goto end;
    }
    for(size_t e=0; e<eventCount; e++) {
        if (!findMagic(f, fileMagicEV, fileName)) {
            goto end;
        }
        DaqEvent daqEvent;
        quint32 eventNumber = 0;
        fread(&eventNumber, sizeof(eventNumber), 1, f);
        daqEvent.eventNumber = eventNumber;
        quint32 masterId = 0;
        fread(&masterId, sizeof(masterId), 1, f);
        daqEvent.masterId = masterId;
        double eventMasterTimeStamp = 0;
        fread(&eventMasterTimeStamp, sizeof(eventMasterTimeStamp), 1, f);
        double eventBField_Gs = 0;
        fread(&eventBField_Gs, sizeof(eventBField_Gs), 1, f);
        daqEvent.BField_Gs = eventBField_Gs;
        DaqEventBoxes daqEventBoxes;
        quint32 boxCount = 0;
        fread(&boxCount, sizeof(boxCount), 1, f);
        for(size_t i=0; i<boxCount; i++) {
            if (!findMagic(f, fileMagicBOX, fileName)) {
                goto end;
            }
            quint32 id = 0;
            fread(&id, sizeof(id), 1, f);
            double eventTimeStamp = 0;
            fread(&eventTimeStamp, sizeof(eventTimeStamp), 1, f);
            double clockFreqFactor = 0;
            fread(&clockFreqFactor, sizeof(clockFreqFactor), 1, f);
            double temp = 0;
            fread(&temp, sizeof(temp), 1, f);
            DaqEventBoxAdcs daqEventBoxAdcs;
            quint32 adcCount = 0;
            fread(&adcCount, sizeof(adcCount), 1, f);
            for(size_t a=0; a<adcCount; a++) {
                if (!findMagic(f, fileMagicADC, fileName)) {
                    goto end;
                }
                quint32 sampleCount = 0;
                fread(&sampleCount, sizeof(sampleCount), 1, f);
                double *samples = new double[sampleCount];
                fread(samples, sizeof(samples[0]), sampleCount, f);
                std::vector<double> daqAdcSamples(sampleCount);
                for(size_t s=0; s<sampleCount; s++)
                    daqAdcSamples[s] = samples[s];
                delete[] samples;
                //    	log_notice("event %zu (%d), box %zu (%u), adc %zu: %zu samples, temp %f", e, eventNumber, i, id, a, sampleCount, temp);
                DaqEventBoxAdc daqEventBoxAdc;
                daqEventBoxAdc.setData(daqAdcSamples);
                daqEventBoxAdcs.push_back(daqEventBoxAdc);
            }
            DaqEventBox daqEventBox;
            daqEventBox.setData(daqEventBoxAdcs);
            daqEventBox.setTemp(temp);
            daqEventBox.setId(id);
            daqEventBox.setClockFreqFactor(clockFreqFactor);
            daqEventBoxes[id] = daqEventBox;
        }
        daqEvent.box = (daqEventBoxes);
        daqEvents.push_back(daqEvent);
//        cerr << daqEvent << endl;
    }
    findMagic(f, fileMagicEOF, fileName);
    end:
    fclose(f);
*/
    return daqEvents;
}
