//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef BPMEVENT_H
#define BPMEVENT_H

#include <QVector>
#include <QMap>
#include <QtGlobal>
#include <QTextStream>
#include <QMetaType>

#include "BpmDataTypes.h"

typedef QVector<quint32> BPMBoxRawSamples;

class BPMBoxEvent {
public:
    BPMBoxEvent(): timeStamp(0), BField_Gs(0), master(false) { }
    friend QTextStream& operator<<(QTextStream &, const BPMBoxEvent &);
    //        void print() const;
    double timeStamp;
    double BField_Gs; // on master only
    bool master;
    BPMBoxRawSamples adc;
};

// all events from single box
class BPMBoxEvents: public QVector<BPMBoxEvent> { };

// all events from all boxes
class BPMMultiBoxEvents: public QMap<int, BPMBoxEvents> {
public:
    BPMMultiBoxEvents() : masterId(0) {}
    void setMasterId(int n) { masterId = n; }
    int getMasterId() const { return masterId; }
    DeviceAttrs attrs;
private:
    int masterId;
};

const double defaultClockFreqFactor = 0.;
/*
class DaqAdcSamples: public myvector<double> {
  public:
        DaqAdcSamples() : clockFreq(defaultClockFreq) {}
        DaqAdcSamples(const myvector<double> &r): myvector<double>(r), clockFreq(defaultClockFreq) {}
//	DaqAdcSamples& operator=(const myvector<double>&);
        void print() const;

  protected:
};
*/
typedef QVector<double> DaqAdcSamples;

class DaqEventBoxAdc {
public:
    DaqEventBoxAdc(): samples(), baseline(0), rms(0), clockFreqFactor(0) { }
    friend QTextStream& operator<<(QTextStream &, const DaqEventBoxAdc &);
    size_t size() const { return samples.size(); }
    //	void resize(size_t n) { samples.resize(n); }
    double operator[](size_t n) const { return samples[n]; }
    void setData(const DaqAdcSamples &r);
    DaqAdcSamples getSamples() const { return samples; }
    double getBaseline() const { return baseline; }
    double getRms() const { return rms; }
    //	std::vector<double> getRms() const;
    //        void print() const;
    //        void printRms() const;
    //        QString getRmsText() const;
    void resample();
    void setClockFreqFactor(double v) { clockFreqFactor = v; }
    double getClockFreqFactor() const { return clockFreqFactor; }
    void resample(double ratio, int algo, double oversample_factor);

protected:
    void updateStats();
    void updateRms();
    DaqAdcSamples samples;
    double baseline;
    double rms;
    double clockFreqFactor;
};

Q_DECLARE_METATYPE(DaqEventBoxAdc);
QDataStream &operator<<(QDataStream &stream, const DaqEventBoxAdc &daqEventBoxAdc);
QDataStream &operator>>(QDataStream &stream, DaqEventBoxAdc &daqEventBoxAdc);

class DaqEventBoxAdcs: public QVector<DaqEventBoxAdc> {
public:
    DaqEventBoxAdcs(size_t n = 0) : QVector<DaqEventBoxAdc>(n) { }
};
typedef int DaqEventBoxId_t;

class DaqEventBox {
public:
    DaqEventBox();
    friend QTextStream& operator<<(QTextStream &, const DaqEventBox &);
    //        void print() const;
    void setData(const DaqEventBoxAdcs &r);
    void setClockFreqFactor(double v);
    void setId(int n) { id = n; }
    double getTemp() const { return temp; }
    void setTemp(double n) { temp = n; }
    void setTimeStamp(double n) { timeStamp = n; }
    double getClockFreqFactor() const { return clockFreqFactor; }
    double getTimeStamp() const { return timeStamp; }
    double getTimeStampCorrected() const { return getTimeStamp() / (1 + getClockFreqFactor()); }
    double getSamplingFreq() const;
    double getSamplingFreqCorrected() const { return getSamplingFreq() / (1 + getClockFreqFactor()); }
    int getId() const { return id; }
    QString getFullName() const;
    size_t size() const { return adc.size(); }
    const DaqEventBoxAdc &operator[](size_t n) const { return adc.at(n); }
    DaqEventBoxAdcs getAdc() const { return adc; }

protected:
    DaqEventBoxId_t id;
    double clockFreqFactor;
    double temp;
    double timeStamp;
    DaqEventBoxAdcs adc;
};

Q_DECLARE_METATYPE(DaqEventBox);
QDataStream &operator<<(QDataStream &stream, const DaqEventBox &daqEventBox);
QDataStream &operator>>(QDataStream &stream, DaqEventBox &daqEventBox);

class DaqEventBoxes: public QMap<int, DaqEventBox> { };


class DaqEvent {
public:
    DaqEvent() : masterId(0), eventNumber(0), BField_Gs(0) { }
    friend QTextStream& operator<<(QTextStream &, const DaqEvent &);
    //        void print() const;
    size_t size() const { return box.size()*4; }
    const DaqEventBoxAdc &operator[](size_t n) const;
    double getMasterTimeStamp() const;
    DaqEventBoxes box;
    int masterId;
    int eventNumber;
    double BField_Gs;
};
Q_DECLARE_METATYPE(DaqEvent);
QDataStream &operator<<(QDataStream &stream, const DaqEvent &daqEvent);
QDataStream &operator>>(QDataStream &stream, DaqEvent &daqEvent);

class AppSettings;
class DaqEvents: public QVector<DaqEvent> {
public:
//    static QString getRawFileNameTimestamped(const time_t &t = 0);
//    static QString writeEvents(const DaqEvents &daqEvents, const QString &fileNname = getRawFileNameTimestamped());
    static QString writeEvents(const AppSettings &settings, const DaqEvents &daqEvents, const QString &fileNname);
    static DaqEvents readEvents(const QString &fileName);
    static bool readRawData(const QString &fileName, DaqEvents *daqEvents, AppSettings *appSettings);
};
Q_DECLARE_METATYPE(DaqEvents);

/*
class BPMEvent {
  public:
        void print() const;
        void setTimestamp(long long v) { ts = v; }
        void setEventNumber(long long v) { en = v; }
        void setChannelData(int channel, const std::vector<double> &d) {data[channel] = d;}
        long long getTimestamp() const {return ts;}
        long long getEventNumber() const {return en;}

  protected:
        double getRms();

    QMap<int, QVector<double> > data;
    long long ts; // timestamp
    long long en; // event number
};
*/
#endif /* BPMEVENT_H */
