//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#define _USE_MATH_DEFINES
#include <cmath>
#include <iostream>
#include <cerrno>
#include <cstdio>
#include <cstring>
#include <ctime>

#include <QDebug>

#include "dsplib/dsplib.h"
#include "dsplib/VectorMaths.h"
#ifdef USE_FITTER
#include "dsplib/fitter.h"
#endif
#include "dsplib/fftengine.h"
#include "dsplib/TimeDomainData.h"
#include "bpmeventanalyser.h"
#include "bpmevent.h"
//#include "bpmbox.h"

// Bandpass FFT Filter - set frequency range
const double lowfreq = 20e3;
const double highfreq = 10.0e6;
//const double highfreq = 50.0e6;



using namespace std;

void normalizePhase(double &phi)
{
    if (phi < -M_PI) phi += 2*M_PI;
    if (phi > M_PI) phi -= 2*M_PI;
}

BeamPosition getBeamPosition(const QVector<FFTSpectrum> &spectrums, const BeamSetup &beamSetup, const double B_Gs, const double timeStamp)
{
    if (spectrums.size() != 4) {
        qWarning() << "Bad number of spectrums: " << spectrums.size();
        return BeamPosition();
    }
    const double kf = beamSetup.kf;
    const double beamCycleFreq = beamSetup.b2freq(B_Gs/1e4)/kf;
    QMap<double, double> cycleFreqWeights;
    cycleFreqWeights[1.0] = 1.0;
    cycleFreqWeights[2.0] = 1.0;
    cycleFreqWeights[3.0] = 1.0;
    cycleFreqWeights[4.0] = 1.0;
    //  cycleFreqWeights[5.0] = 1.0;
    //  cycleFreqWeights[1.0 * kf] = 0.0;
    //  cycleFreqWeights[2.0 * kf] = 0.0;
    //  std::cout << "===============================================================" << std::endl;
    //  std::cout << "B_Gs="<<B_Gs <<", BeamCycleFreq=" << beamCycleFreq/1e3 << " kHz" << std::endl;
    /*
  for(int c=0; c<1; c++) {
        const FFTSpectrum &spectrum = spectrums[c];
    std::cout << "Spectrum["<<c<<"]:" << std::endl;
    for(size_t i=770; i<790; i++)
          std::cout << "FFTSpectrum[" << i << "]: freq " << spectrum[i].freq << ", power " << val2db(spectrum[i].power/8192.) << " dB, phase " << spectrum[i].phase << std::endl;
  }
*/
    QMap<double, QVector<double> > harmonicPower;
    QMap<double, QVector<double> > harmonicPowerErr;
    QMap<double, QVector<double> > harmonicPhase;
    QMap<double, double> harmonicWeight;
    Q_FOREACH(double harmonicNumber, cycleFreqWeights.keys()) {
        double weight = cycleFreqWeights[harmonicNumber];
        double freq = beamCycleFreq * harmonicNumber;
        QVector<double> powers(4);
        QVector<double> powerErrs(4);
        QVector<double> phases(4);
        for(int c=0; c<4; c++) {
            const FFTSpectrum &spectrum = spectrums[c];
            const FFTSpectrumElement &fftApprox = spectrum.getApproximation(freq);
            powers[c] = fftApprox.power;
            powerErrs[c] = fftApprox.powerError;
            phases[c] = fftApprox.phase;
        }
        harmonicPower[harmonicNumber] = powers;
        harmonicPowerErr[harmonicNumber] = powerErrs;
        harmonicPhase[harmonicNumber] = phases;
        harmonicWeight[harmonicNumber] = weight;
    }
    //  std::cout << "***************************************************************" << std::endl;
    BeamPosition beamPos;
    beamPos.B_Gs = B_Gs;
    beamPos.time = timeStamp;
    size_t count = 0;
    Q_FOREACH(double harmonicNumber, harmonicPower.keys()) {
        const QVector<double> &p = harmonicPower[harmonicNumber];
        const QVector<double> &pe = harmonicPowerErr[harmonicNumber];
        const QVector<double> &phase = harmonicPhase[harmonicNumber];
        const double weight = harmonicWeight[harmonicNumber];
        double posH = p[0]?(p[1]/p[0]):0;
        double posV = p[2]?(p[3]/p[2]):0;
        double phaseH = phase[1]-phase[0];
        double phaseV = phase[3]-phase[2];
        normalizePhase(phaseH);
        normalizePhase(phaseV);
        int signH = (phaseH >= -M_PI/2 && phaseH < M_PI/2)?1:-1;
        int signV = (phaseV >= -M_PI/2 && phaseV < M_PI/2)?1:-1;
        posH *= signH;
        posV *= signV;
        beamPos.posH += posH*weight;
        beamPos.posV += posV*weight;
        beamPos.powerH += p[0];
        beamPos.powerV += p[2];
        beamPos.powerErrH += pe[0];
        beamPos.powerErrV += pe[2];
        count++;
        //	std::cout << "f: " << beamCycleFreq*harmonicNumber/1e3 << " kHz, H: " << posH*200. << "%, V: " << posV*200. << "%" << std::endl;
        /*
        double powerdB[4];
        for(int c=0; c<4; c++)
          powerdB[c] = val2db(p[c] / 8192.);
//	std::cout << "f: " << beamCycleFreq*harmonicNumber/1e3 << " kHz, H: " << posH*200. << "%, phase: " << phaseH << std::endl;
        std::cout << "f: " << beamCycleFreq*harmonicNumber/1e3 << " kHz"
        <<", p0: " << powerdB[0] << ", p1: " << powerdB[1]
        << ", phi0: " << phase[0] << "phi1: " << phase[1]
        <<", p2: " << powerdB[2] << ", p3: " << powerdB[3]
        << ", phi2: " << phase[2] << "phi3: " << phase[3]
         << std::endl;
*/
    }
    if (count>0) {
        beamPos.posH /= count;
        beamPos.posV /= count;
        beamPos.powerErrH /= count;
        beamPos.powerErrV /= count;
        beamPos.powerH /= count;
        beamPos.powerV /= count;
    }
    //  std::cout << "H: " << beamPos.powerHdB() << " dB, " << beamPos.posH*100. << "%, "
    //  << "V: " <<  beamPos.powerVdB() << " dB, " << beamPos.posV*100. << "%" << std::endl;
    //  std::cout << "===============================================================" << std::endl;
    return beamPos;
}

BpmEventAnalyser::BpmEventAnalyser(const BeamSetup &r)
    : beamSetup(r),
    fft(new RFFTEngine())
#ifdef USE_FITTER
    , fitter(new FitterSine())
#endif
{
}

BpmEventAnalyser::~BpmEventAnalyser()
{
#ifdef USE_FITTER
    delete fitter;
#endif
    delete fft;
}

void BpmEventAnalyser::clearSpectraSum()
{
    fftSpectraSum.clear();
    fftSpectraCount.clear();
}

void BpmEventAnalyser::clearMeasurements()
{
    beamPositions.clear();
}

void BpmEventAnalyser::addEvents(const DaqEvents &events)
{
    Q_FOREACH(DaqEvent event, events) {
        addEvent(event);
    }
}

#ifdef USE_FITTER
FreqMap BpmEventAnalyser::measureFrequency(const DaqEvents &events) const
{
    FreqMap m;
    foreach_const(events, ei) {
        const DaqEvent &event = (*ei);
        const DaqEventBoxes &boxes = event.box;
        foreach_const(boxes, bi) {
            int id = (*bi).first;
            const DaqEventBox &box = (*bi).second;
            for(size_t c=0; c<box.size(); c++) {
                const DaqEventBoxAdc &adc = box[c];
                const QVector<double> &samples = adc.getSamples();

                if (id == 2 && (c == 0 || c == 2)) {
                    const QVector<double> &vx = samples;
                    //  const double *x = &vx[0]; // dirty hack
                    size_t len = vx.size();
                    //  int len = std::min(vx.size(), 1024u);
                    double *x = new double[len];
                    for(size_t i=0; i<len; i++)
                        x[i] = vx[i];
                    int fixpar = 0;
                    double par[5] = {0,0,0,0,0};
                    fitstatus_t fitstatus;
                    fitstatus.niter = 10;
                    fitstatus.fnorm = 0;
                    fitter->fit (fixpar, x, len/2, par, &fitstatus);
                    // f(x) = p0 + p1 * sin(p2 + p3*(1+1e-6*p4)*x)
                    //  printf("fit result: f(x) = %e + %e * sin(%e + %e * x)\n", par[0], par[1], par[2], par[3]);
                    double f1 = fabs(par[3]/(2*M_PI));
                    fitter->fit (fixpar, &x[len/2], len/2, par, &fitstatus);
                    double f2 = fabs(par[3]/(2*M_PI));

                    double fc = (f1+f2)/2;
                    double f0 = fc - (f2-f1);
                    double measuredSineFreq = f0 * BPMBox::sampleFreq();
                    FreqMapEntry me;
                    me.time = event.getMasterTimeStamp();
                    me.field = event.BField_Gs;
                    me.freq = measuredSineFreq;
                    me.freqVary = (f2-f1)/(f2+f1);
                    m[c].push_back(me);
                    delete [] x;
                }

            }
        }
    }
    return m;
}
#endif
void BpmEventAnalyser::addEvent(const DaqEvent &event)
{
    //  event.print();
    size_t eventNumber = event.eventNumber;
    double eventMasterTimeStamp = event.getMasterTimeStamp();
    const DaqEventBoxes &boxes = event.box;
    Q_FOREACH(int id, boxes.keys()) {
        //	  if (id != 3) continue;
        const DaqEventBox &box = boxes[id];
        double eventTimeStamp = box.getTimeStamp();
        double samplingFreq = box.getSamplingFreq();
        double timeStampDiff = eventTimeStamp-eventMasterTimeStamp;
        if (fabs(timeStampDiff) > 1e-4) {
            qWarning() << "ev " << eventNumber << ", " << box.getFullName() << ": BAD timeStampDiff " << timeStampDiff;
        }
        QVector<FFTSpectrum> FFTSpectrums;
        QVector<int> rmsValues(box.size());
        for(size_t c=0; c<box.size(); c++) {
            const DaqEventBoxAdc &adc = box[c];
            const DaqAdcSamples &daqAdcSamples = adc.getSamples();
            rmsValues[c] = VectorMaths::calculateRMS(daqAdcSamples.toStdVector());//.rms();
            TimeDomainData samples(daqAdcSamples, eventTimeStamp, samplingFreq);
            //		samples.print();
            //                int channelIndex = BPMBox::getChannelIndex(id, c);
            //                const FFTSpectrum &spectrum = fft->getFFTSpectrum(samples);

            size_t lowbin = 0;
            size_t highbin = 0;
            FFTSpectrum filteredSpectrum;
            {
                const FFTSpectrum &spectrum = fft->getFFTSpectrum(samples);
                highbin = spectrum.size();
                for(int i=0; i<spectrum.size(); i++) {
                    double freq = spectrum[i].freq;
                    if (freq <= lowfreq) lowbin = i;
                    if (freq >= highfreq) {
                        highbin = i;
                        break;
                    }
                }
                //                    qDebug() << "Bandpass filter: bins " << lowbin << highbin;
                for(size_t i=lowbin; i<highbin; i++) {
                    filteredSpectrum.push_back(spectrum[i]);
                }
            }
            FFTSpectrums.push_back(filteredSpectrum);

            /*
if (id == 2 && c == 0) {
  const QVector<double> &vx = samples.v;
  const double *x = &vx[0]; // dirty hack
  int len = std::min(vx.size(), MAX_FITSINE_POINTS);
  int fixpar = 0;
  double par0[4] = {0,0,0,0};
  double par[4] = {0,0,0,0};
  fitstatus_t fitstatus;
  fitstatus.niter = 100;
  fitstatus.fnorm = 0;
  fitter->fit (fixpar, x, len, par0, par, &fitstatus);
// f(x) = p0 + p1 * sin(p2 + p3*x)
  printf("fit result: f(x) = %e + %e * sin(%e + %e * x)\n", par[0], par[1], par[2], par[3]);
  double measuredSineFreq = par[3]/(2*M_PI) * BPMBox::sampleFreq();
  filedFreqMap[event.BField_Gs] = measuredSineFreq;
}
*/
            /*
                QVector<FFTSpectrumPeak> spectrumPeaks = spectrum.getPeaks();
*/
//            size_t newsize = std::max(fftSpectraSum[id][c].size(), filteredSpectrum.size());
            //		spectrum.print();
            fftSpectraSum[id][c] += filteredSpectrum; fftSpectraCount[id][c]++;
            //		fftSpectraSum[id][c] = spectrum; fftSpectraCount[id][c] = 1;// FIXME: no spectrum addition

            /*
                fftSpectrum[id][c].resize(newsize); // extend
                for(size_t i=0; i<vp.size(); i++) {
                  fftPowDBFS[id][c].v[i] += vp[i];
                }
*/
            //		fftPowDBFS[id][c].n ++;
        }
        BeamPosition beamPos = getBeamPosition(FFTSpectrums, beamSetup, event.BField_Gs, event.getMasterTimeStamp());
        beamPos.id = id;
        if (rmsValues.size() > 0)
            beamPos.rmsH = rmsValues.at(0);
        if (rmsValues.size() > 2)
            beamPos.rmsV = rmsValues.at(2);
        beamPositions[id].push_back(beamPos);
    }
    //  appendBeamPosition(beamPositions, "/data/beampos.txt");
}

BoxAdcFFTSpectra BpmEventAnalyser::getFFTSpectra() const
{
    BoxAdcFFTSpectra spectra;
    Q_FOREACH(int box_id, fftSpectraSum.keys()) {
        const AdcFFTSpectra &adcSpectra = fftSpectraSum[box_id];
        Q_FOREACH(int adc_id, adcSpectra.keys()) {
            const FFTSpectrum &spectrum = adcSpectra[adc_id];
            Q_ASSERT(fftSpectraCount.contains(box_id));
            const BoxFftSpectraCount &boxFftSpectraCount = fftSpectraCount[box_id];
            Q_ASSERT(boxFftSpectraCount.contains(adc_id));
            quint64 count = boxFftSpectraCount[adc_id];
            spectra[box_id][adc_id] = spectrum / (count?count:1);
        }
    }
    return spectra;
    //  return fftSpectraSum;
}

void BpmEventAnalyser::saveFieldFreqMap(const FreqMap &m, const QString &fileName) const
{
    /*
  if (m.empty()) return;
  FILE *f;
  f = fopen(fileName.c_str(), "w");
  if (!f) {
    using namespace std;
    cerr << "Can't open file " << fileName << ": " << strerror(errno) << endl;
    return;
  }
  foreach_const(m, it) { // FIXME
        double time = (*it).time;
        double bfield = (*it).field;
        double freq = (*it).freq;
        double freqVary = (*it).freqVary;
        double refFreq = beamSetup.b2freq(bfield/1e4);
        fprintf(f, "%e\t%e\t%e\t%e\t%e\n", time, bfield, freq, (freq-refFreq)/refFreq, freqVary);
  }
  fclose(f);
*/
}

