//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <iostream>
#include <algorithm>
#include <cmath>

#include "FFTSpectrum.h"

std::ostream &operator<<(std::ostream &stream, const FFTSpectrum &r)
{
    stream << "FFTSpectrum: " << r.size() << " bins" << std::endl;
    for(size_t i=0; i<std::min(size_t(10), r.size()); i++)
        stream << "FFTSpectrum[" << i << "]: freq " << r[i].freq
               << ", power " << r[i].power
               << ", phase " << r[i].phase << std::endl;
    return stream;
}

void FFTSpectrum::print() const
{
    std::cout << (*this);
}

double FFTSpectrum::getPeak() const
{
    return (*std::max_element(begin(), end())).power;
}

double FFTSpectrum::rms() const
{
    double v = 0;
    for(size_t i=1; i<size(); i++)
        v += (*this)[i].power;
    return v;
}

FFTSpectrumPeakVector FFTSpectrum::getPeaks(double threshold) const
{
    FFTSpectrumPeakVector peaks;
    if (size() <= 2) return peaks;
    // find absolute maximum
    double maxmag = getPeak();
    // find local maximum
    double vthr = maxmag*threshold;
    for(size_t i=1; i<size()-1; i++)
        if ((*this)[i].power - 0.5*( (*this)[i-1].power+ (*this)[i+1].power) >= vthr) {
        FFTSpectrumPeak p;
        p.freq = (*this)[i].freq; // FIXME - intrepolate
        p.power = (*this)[i].power; // FIXME
        peaks.push_back(p);
    }
    std::sort(peaks.begin(), peaks.end(), FFTSpectrumPeakLt);
    size_t n=0;
    for(FFTSpectrumPeakVector::const_iterator it = peaks.begin(); it != peaks.end(); ++it) {
        double freq = (*it).freq;
        double power = (*it).power;
        std::cout << "Spectrum peak "<<n<<": freq " << freq << ", power " << power << std::endl;
        n++;
        if (n>10) {
            std::cout << "More peaks follow... " << std::endl;
            break;
        }
    }
    return peaks;
}

FFTSpectrum FFTSpectrum::operator+(const FFTSpectrum &v) const
{
    FFTSpectrum r(*this);
    const size_t newsize = std::max(r.size(), v.size());
    r.resize(newsize);
    for(size_t i=0; i<v.size(); i++) {
        if (r[i].freq != v[i].freq) {} // FIXME
        r[i].freq = v[i].freq; // overwrite
        r[i].power += v[i].power;
        r[i].phase += v[i].phase; // ?
    }
    return r;
}

FFTSpectrum FFTSpectrum::operator/(double n) const
{
    FFTSpectrum r(*this);
    for(size_t i=0; i<r.size(); i++) {
        r[i].power /= n;
        r[i].phase /= n;
    }
    return r;
}

FFTSpectrumElement FFTSpectrum::getApproximation(double frequency) const
{
    FFTSpectrumElement elem;
    if (empty()) return elem;
    size_t nearestBin = 0;
    double nearestFreqDistance = fabs(at(0).freq-frequency);
    for(size_t i=0; i<size(); i++) {
        double dist = fabs(at(i).freq - frequency);
        if (dist < nearestFreqDistance) {
            nearestFreqDistance = dist;
            nearestBin = i;
        }
    }
    // FIXME: assume frequency is monotonic over spectrum elements
    if (nearestBin < 1) return at(0);
    if (nearestBin >= size())
        return at(size()-1);

  elem = at(nearestBin);
  size_t b0 = nearestBin-1;
  size_t b1 = nearestBin;
  size_t b2 = nearestBin+1;
  double f0 = at(b0).freq;
  double f1 = at(b1).freq;
  double f2 = at(b2).freq;
  double w0 = at(b0).power;
  double w1 = at(b1).power;
  double w2 = at(b2).power;
/*
  elem.freq = (f0+f1+f2)/3;
  elem.power = (w0+w1+w2)/3;
  elem.phase = (p0+p1+p2)/3;
*/
  /*
    // parabolic peak estimation
    // assume equally spaced frequency bins: f2-f1 = f1-f0
    double p = 0.5 * (w0 - w2) / (w0 - 2.0 * w1 + w2);
    elem.freq = f1 + (f2-f0)/2 * p;
    elem.power = w1 - 0.25*p*(w2-w0);
    elem.phase = at(b1).phase;
    double freqError = (elem.freq - f1) / f1;
    double magnError = (elem.power - w1) / w1;
    std::cout << "peak estimation: freqError = " << freqError << ", magnError = " << magnError << std::endl;
*/
  // error estimation - use background
  if (b1 >= 2 && b1+2 < size()) {
      double bgr = (at(b1-2).power + at(b1+2).power) / 2;
      elem.powerError = bgr;
  }
    return elem;
}
