//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>

// Compile this source with gcc -fast-math option

#include "dsplib.h"

//void LinearFit(double x[], double y[], int n, double *a, double *b)
void LinearFit (const double *x, const double *y, int n, double *a, double *b, double *c)
{

    // Find a,b for array of (x,y) of length n
    // y = ax + b

    int i;
    double xa = 0, ya = 0, stt = 0, xy = 0, aa, bb, chi2 = 0;

    if (n < 1)
        return;

    for (i = 0; i < n; i++) {
        xa += x[i];
        ya += y[i];
        //    printf("%d: %f, %f\n", i, x[i], y[i]);
    }
    xa /= n;
    ya /= n;
    for (i = 0; i < n; i++)
        stt += sqr (x[i] - xa);
    for (i = 0; i < n; i++)
        xy += (x[i] - xa) * y[i];
    if (stt != 0)
        bb = xy / stt;
    else
        bb = 0;
    aa = ya - xa * bb;
    for (i = 0; i < n; i++)
        chi2 += sqr (y[i] - aa - bb * x[i]);
    //  printf("Line fit: a=%f, b=%f, chi2=%f\n", aa, bb, chi2);
    if (a)
        *a = bb;
    if (b)
        *b = aa;
    if (c)
        *c = chi2;
}

void LinearFitInt (const int *y, int n, double *a, double *b, double *c)
{

    // Find a,b for array of (x,y) of length n
    // y = ax + b

    int i;
    double ya = 0, stt = 0, xy = 0, aa, bb, chi2 = 0;

    if (n < 1)
        return;

    for (i = 0; i < n; i++) {
        ya += y[i];
    }
    ya /= n;
    for (i = 0; i < n; i++)
        stt += sqr (i - 1);
    for (i = 0; i < n; i++)
        xy += (i - 1) * y[i];
    if (stt != 0)
        bb = xy / stt;
    else
        bb = 0;
    aa = ya - bb;
    for (i = 0; i < n; i++)
        chi2 += sqr (y[i] - aa - bb * i);
    printf ("Line fit int: a=%f, b=%f, chi2=%f\n", aa, bb, chi2);
    if (a)
        *a = bb;
    if (b)
        *b = aa;
    if (c)
        *c = chi2;
}

void convolve (const int *x1, int l1, const int *x2, int l2, int *out)
{

    // Integer Convolution

    int i, j;

    memset(out, 0, sizeof(int)*(l1+l2));
    //  for (i = 0; i < l1 + l2; i++) out[i] = 0;

    for (i = 0; i < l1; i++) {
        for (j = 0; j < l2; j++)
        {
            out[i + j] += x1[i] * x2[j];
        }
    }
    /*
  for (i = 1 + fir_delay; i < l1 + l2 - 1; i++)
    out[i - fir_delay] = out[i];
  for (i = l1 + l2 - fir_delay; i < l1 + l2 - 1; i++)
    out[i - fir_delay] = 0;
*/
}

void convolve_i_s_i (const int *x1, int l1, const short int *x2, int l2, int *out)
{

    // Integer Convolution

    //  int test[1024];

    //  int i;

    // Important! Zero buffer. Buffer is durty on enter.
    // memset is fastest
    memset(out, 0, sizeof(int)*(l1+l2));
    //  int *optr = &out[0];
    //  for (i = 0; i < l1 + l2; i++) *optr++ = 0;

    /*
// This is trivial implementation. Demo.

  {
  int i, j;
  for (i = 0; i < l1; i++)
    for (j = 0; j < l2; j++)
    {
      out[i + j] += x1[i] * x2[j];
    }
  }
*/
    /*
  memset(test, 0, sizeof(int)*(l1+l2));
// Good code for continuous filtering with head overlap, assuming missing samples = 0
  {
  const int *coef = x1;
  int nk = l1;
  const short int *data = x2;
  int nd = l2;
  int k, j;
  assert(nk<=nd);

// Compute head overlap
//  test[0] = coef[0] * data[0];
//  test[1] = coef[1] * data[0]+coef[0]*data[1];
//  test[2] = coef[2] * data[0]+coef[1]*data[1]+coef[0]*data[2];
//  test[nk-2] = ;
  for (k = 0; k < nk-1; k++) {
  {
    int tmp=0;
    for(j=0;j<=k;j++) {
      tmp += coef[k-j] * data[j];
    }
    test[k] = tmp;
  }
  }
// Compute main part
  for (j = 0; j < nd; j++) {
    int tmp = 0;
    for (k = 0; k < nk; k++)
    {
      tmp += coef[nk-1-k] * data[j+k];
    }
    test[nk-1+j] = tmp;

  }

  }
*/

    // Welcome to real life. This algo is two times faster.
    // Avoid index operations. Use pointers.
    {
        int i;
        for (i = 0; i < l1; i++)
        {
            int *optr = &out[i];
            int x1i = x1[i];
            const short int *x2ptr = x2;
            int j;
            for (j = 0; j < l2; j++)
            {
                *optr++ += x1i * *x2ptr++;
            }
        }
    }

    /*
// Another solution. Result is short, cannot use it.
  {
    int y_ptr, c_ptr, x_ptr;
    int n, k;
    register int temp;
    y_ptr = 0;
    for (n = 0; n < l2; n++) {
      c_ptr = 0;
      x_ptr = n;
      temp = 0;
      for (k = 0; k < l1; k++)
	    temp += x1[l1-1-c_ptr++] * x2[x_ptr++];
      test[l1-1+y_ptr++] = temp;
    }
  }
*/
    /*
  {
    int i;
    printf("l1=%2d, l2=%2d, result %d\n", l1, l2, l1+l2);
    for (i = 0; i < l1 + l2; i++)
//    if (out[i] != test[i])
      printf("[%2d]: int %8d out %8d test %8d\n", i, x2[i], out[i], test[i]);
  }
*/
    /*
// code above optimized
  {
    int *y_ptr;
    const int *c_ptr;
    const short int *x_ptr;
    register int temp;
    int n, k;
    y_ptr = &out[0];
    for (n = 0; n < l2; n++) {
      c_ptr = &x1[0];
      x_ptr = &x2[n];
      temp = *c_ptr++ * *x_ptr++;
      for (k = 1; k < l1; k++)
	    temp = temp + *c_ptr++ * *x_ptr++;
      *y_ptr++ = temp;
    }
  }
*/
    //  for (i = 0; i < l1 + l2; i++) out[i] = -out[i];

    /*
  for (i = 1 + fir_delay; i < l1 + l2 - 1; i++)
    out[i - fir_delay] = out[i];
  for (i = l1 + l2 - fir_delay; i < l1 + l2 - 1; i++)
    out[i - fir_delay] = 0;
*/
}

void convolve_id (const int *x1, int l1, const double *x2, int l2, double *out)
{

    // Integer+Double Convolution

    int i, j;

    for (i = 0; i < l1 + l2; i++)
        out[i] = 0;

    for (i = 0; i < l1; i++)
        for (j = 0; j < l2; j++) {
        out[i + j] += x1[i] * x2[j];
    }
}

void convolve_d (const double *x1, int l1, const double *x2, int l2, double *out)
{

    // Double Convolution

    int i, j;

    for (i = 0; i < l1 + l2; i++)
        out[i] = 0;

    for (i = 0; i < l1; i++)
        for (j = 0; j < l2; j++) {
        out[i + j] += x1[i] * x2[j];
    }
}

void deriv2 (const int *x, int len, int *out)
{
    int i;
    for (i = 1; i < len - 1; i++)
        out[i] = 2 * x[i] - x[i - 1] - x[i + 1];
}

double gauss (double x, double sigma)
{

    // An unnormalized gaussian function with mean=0 and sigma
    // Divide result by sqrt(2*Pi) = 2.50662827463100024 if normalized output needed

    double arg, res;
    // Calculate a gaussian function with mean and sigma.
    if (sigma == 0)
        return 1.e30;
    arg = x / sigma;
    res = exp (-0.5 * arg * arg);
    return res;
    //   return res/(2.50662827463100024*sigma); //sqrt(2*Pi)=2.50662827463100024
}

// Windowing functions
// i = 0, 1, ..., m

double wf_barthann (double i, int m)
{
    double arg = i / m - .5;
    double parg = arg;
    if (parg < 0)
        parg = -parg;
    return (0.62 - 0.48 * parg + 0.38 * cos (2 * pi * arg));
}

double wf_barlett (double i, int m)
{
    if (m <= 0)
        return (1);
    if ((i < 0) || (i > m))
        return (0);
    if (i < .5 * m)
        return (2. * i / m);
    else
        return (2. - 2. * i / m);
}

double wf_blackman (double i, int m)
{
    return (0.42 - 0.5 * cos (2 * pi * i / m) + 0.08 * cos (4 * pi * i / m));
}

double wf_blackmanharris (double i, int m)
{
    // returns an (m+1)-point, minimum , 4-term Blackman-harris window.
    // The window is minimum in the sense that its maximum sidelobes are minimized.
    double arg = 2 * pi * i / m;
    return (0.35875 - 0.48829 * cos (arg) + 0.14128 * cos (2 * arg) - 0.01168 * cos (3 * arg));
}

double wf_bohman (double i, int m)
{
    double arg = 2 * i / m - 1;
    double parg = arg;

    if ((i < 0) || (i > m))
        return (0);
    if (parg < 0)
        parg = -parg;
    return ((1 - parg) * cos (pi * parg) + 1 / pi * sin (pi * parg));
}

double wf_flattop (double i, int m)
{
    double arg = 2 * pi * i / m;
    return (1 - 1.93 * cos (arg) + 1.29 * cos (2 * arg) - 0.388 * cos (3 * arg) + 0.0322 * cos (4 * arg));
}

double wf_gauss (double i, int m, double r)
{
    double x = i / m - .5;
    return (gauss (x, r));
}
double wf_gauss_075 (double i, int m)
{
    return (wf_gauss (i, m, .75));
}
double wf_gauss_050 (double i, int m)
{
    return (wf_gauss (i, m, .50));
}
double wf_gauss_025 (double i, int m)
{
    return (wf_gauss (i, m, .25));
}

double wf_hamming (double i, int m)
{
    return (0.53836 - 0.46164 * cos (2 * pi * i / m));
}

double wf_hann (double i, int m)
{
    return (0.5 * (1 - cos (2 * pi * i / m)));
}

double wf_nuttall (double i, int m)
{
    // Similar to BlackmanHarris, but produce slightly lower sidelobes.
    double arg = 2 * pi * i / m;
    return (0.3635819 - 0.4891775 * cos (arg) + 0.1365995 * cos (2 * arg) - 0.0106411 * cos (3 * arg));
}

double wf_rect (double i, int m)
{
    return (1);
}

double wf_triang (double i, int m)
{
    return (wf_barlett (i + 1, m + 2));
}

double wf_tukey (double i, int m, double r)
{
    double j = i - .5 * m;
    double pj = j;

    if (r <= 0)
        return (1);
    if (r >= 1)
        r = 1;

    if (pj < 0)
        pj = 0. - pj;
    if (pj < (r * .5 * m))
        return (1);
    return (cos (pi * (pj - r * .5 * m) / ((1 - r) * m)));

}
double wf_tukey_075 (double i, int m)
{
    return (wf_tukey (i, m, .75));
}
double wf_tukey_050 (double i, int m)
{
    return (wf_tukey (i, m, .50));
}
double wf_tukey_025 (double i, int m)
{
    return (wf_tukey (i, m, .25));
}

double windowingFunction (fwinf_t win_type, double i, int m)
{
    switch (win_type) {
    case WF_BARTHANN: return wf_barthann(i, m);
    case WF_BARLETT: return wf_barlett(i, m);
    case WF_BLACKMAN: return wf_blackman(i, m);
    case WF_BLACKMANHARRIS: return wf_blackmanharris(i, m);
    case WF_BOHMAN: return wf_bohman(i, m);
    case WF_FLATTOP: return wf_flattop(i, m);
    case WF_GAUSS_075: return wf_gauss(i, m, 0.75);
    case WF_GAUSS_050: return wf_gauss(i, m, 0.50);
    case WF_GAUSS_025: return wf_gauss(i, m, 0.25);
    case WF_HAMMING: return wf_hamming(i, m);
    case WF_HANN: return wf_hann(i, m);
    case WF_NUTTALL: return wf_nuttall(i, m);
    case WF_RECT: return 1;
    case WF_TRIANG: return wf_triang(i, m);
    case WF_TUKEY_075: return wf_tukey(i, m, 0.75);
    case WF_TUKEY_050: return wf_tukey(i, m, 0.50);
    case WF_TUKEY_025: return wf_tukey(i, m, 0.25);
    default: return 1;
    }
}

void Convolve (const int *x1, int l1, const int *x2, int l2, int *out)
{

    // Integer Convolution

    int i, j, lout;

    lout = l1 + l2;

    for (i = 0; i < lout; i++)
        out[i] = 0;

    for (i = 0; i < l1; i++)
        for (j = 0; j < l2; j++) {
        out[i + j] += x1[i] * x2[j];
    }
}

double fft_noisefloor (double rmsnoise, int rank)
{
    return (20. * log (rmsnoise / sqrt (1. * rank)) / log (10.));
}

void DFT (const double *x, int n, double *S, double *C)
{

    // Discrete Fourier Transform, order n
    // Real form

    int i, k;
    double arg, is, ic;

    for (k = 0; k < (n / 2 + 1); k++) {
        S[k] = 0;
        C[k] = 0;
        for (i = 0; i < n; i++) {
            arg = 2 * pi * k * i / n;
            is = sin(arg);
            ic = cos(arg);
            S[k] += is * x[i];
            C[k] += ic * x[i];
        }
        S[k] /= sqrt (n / 2 + .5);
        C[k] /= sqrt (n / 2 + .5);
    }
    //  if (n%2 == 0) /* N is even */ C[n/2] *= .5; /* Nyquist freq. */
}

void IDFT (double *x, int n, const double *S, const double *C)
{

    // Inverse Discrete Fourier Transform, order n
    // Real form

    int i, k;
    double c, arg, is, ic;

    for (k = 0; k < n; k++) {
        x[k] = 0;
        for (i = 0; i < (n / 2 + 1); i++) {
            c = C[i];
            if ((i == 0) || (i == n / 2))
                c *= 0.5;
            arg = 2 * pi * i * k / n;
            is = sin(arg);
            ic = cos(arg);
            x[k] += (is * S[i] + ic * c);
            //      x[k] += (sin(2*pi*i*k/n)*S[i] + cos(2*pi*i*k/n)*c);
        }
        x[k] /= sqrt (n / 2 + .5);
    }
}

void RECT2POW (const double *S, const double *C, int n, double *power)
{

    // Calculate power from rectangular FFT

    int i;
    for (i = 0; i < n; i++)
        power[i] = hypot (S[i], C[i]);
}

void RECT2PHA (const double *S, const double *C, int n, double *phase)
{

    // Calculate phase from rectangular FFT

    int i;
    for (i = 0; i < n; i++)
        phase[i] = atan2 (S[i], C[i]);
}

void Compdiv (double a, double b, double c, double d, double *e, double *f)
{

    // Complex Divide operation in rectangular form:
    // (e+if)=(a+ib)/(c+id)

    double mod2;
    if (!e || !f)
        return;
    //  mod2 = c*c+d*d;
    mod2 = hypot2 (c, d);
    if (mod2 == 0) {
        *e = 1.;
        *f = 1.;
        return;
    }
    *e = (a * c + b * d) / mod2;
    *f = (b * c - a * d) / mod2;
}

double detpulse (double a, double t, double rt, double ft)
{

    // Simulate detector pulse
    //

    double v, r, f, e;

    //  if (rt > ft) m = ft; else m = rt;

    //  e = TMath::Log(ADC_RES*10); // set error to 0.1 resolution
    e = log (1e6);

    r = t / rt;
    f = t / ft;
    if (f > e)
        return (0);
    if (r < -4)
        return (0);

#ifndef _MSC_VER
    v = a * (erf (r) + 1) / 2 * exp (-f);
#else
    if (f>0)
        v = a * exp (-f);
    else
        v = 0;
#endif   // if (v<0) v=0;

    return v;
}

void Normalize (double *x, int n)
{

    // Normalize to integral=1

    int i;
    double sum;
    sum = 0;
    for (i = 0; i < n; i++)
        sum += x[i];
    if (sum != 0)
        for (i = 0; i < n; i++)
            x[i] /= sum;
}

void Make_Discrete (const double *x, int *out, int n, int bits)
{

    // Scale to fit in <bits>-bit integer

    int i, imin, imax;
    double min, max, scalepos, scaleneg, scale;

    if (bits <= 1)
        return;
    imin = BITS2MININT (bits) /* - (1<<(bits-1)) */ ;
    imax = BITS2MAXINT (bits) /* ((1<<(bits-1)) - 1) */ ;
    min = 0;
    max = 0;

    for (i = 0; i < n; i++) {
        if (x[i] > max)
            max = x[i];
        if (x[i] < min)
            min = x[i];
    }
    if (max > 0)
        scalepos = imax / max;
    else
        scalepos = 1e38;
    if (min < 0)
        scaleneg = imin / min;
    else
        scaleneg = 1e38;
    if (scalepos < scaleneg)
        scale = scalepos;
    else
        scale = scaleneg;
    // printf("Make_Discrete(): min=%f, max=%f, scaling by %f (neg %f, pos %f)\n", min, max, scale, scaleneg, scalepos);
    if (scale != 0)
        for (i = 0; i < n; i++)
            out[i] = (int) (x[i] * scale);
}

int intsum (const int *x, int len)
{
    int i, s = 0;

    for (i = 0; i < len; i++)
        s += x[i];
    return (s);
}

// Max/Min value of array elements (int)
double maxdbl (const double *x, int len)
{
    int i;
    double v = -1e37;
    for (i = 0; i < len; i++)
        if (x[i] > v)
            v = x[i];
    return (v);
}
double mindbl (const double *x, int len)
{
    int i;
    double v = 1e37;
    for (i = 0; i < len; i++)
        if (x[i] < v)
            v = x[i];
    return (v);
}

// DC value
double dcdbl (const double *x, int len)
{
    int i;
    double v = 0;

    if (len <= 0)
        return (0);
    for (i = 0; i < len; i++)
        v += x[i];
    v /= len;
    return (v);
}

// RMS value
double rmsdbl (const double *x, int len, double dc)
{
    int i;
    double v = 0;

    if (len <= 0)
        return (0);
    for (i = 0; i < len; i++)
        v += sqr (x[i] - dc);
    v /= len;
    return (v);
}

// Decibels to value conversion
double db2val (double db)
{
    double value;

    value = exp (db / 20. * log (10.));
    return (value);
}

// Value to decibels conversion
double val2db (double val)
{
    double db;

    db = 20 * (log(val) / log (10.));
    return (db);
}

//double rndm ()
//{
//// System buil-in
//// Random number in range [0,1) with period approx. 16*((2**31)-1)
//  double value;
//  value = (double) rand () / RAND_MAX;
//  return (value);
//}

//unsigned int f1Seed;

//double rndm1 ()
//{
////  Machine independent random number generator.
////  Produces uniformly-distributed floating points between 0 and 1.
////  Identical sequence on all machines of >= 32 bits.
////  Periodicity = 10**8
////  Universal version (Fred James 1985).
////  generates a number in ]0,1]

//  const double kCONS = 4.6566128730774E-10;
//  const int kMASK24 = 2147483392;
//  unsigned int jy;

//  f1Seed *= 69069;
//  jy = (f1Seed & kMASK24);      // Set lower 8 bits to zero to assure exact float
//  if (jy)
//    return kCONS * jy;
//  return rndm ();
//}



//int fSeed1 = 9876;
//int fSeed2 = 54321;

//double rndm2 ()
//{
////  Machine independent random number generator.
////  Produces uniformly-distributed floating points in [0,1]
////  Identical sequence on all machines of >= 32 bits.
////  Periodicity > 10**14

//  double iz, r;
//  int k = (int) (fSeed1 / 53668);
//  fSeed1 = 40014 * (fSeed1 - k * 53668) - k * 12211;
//  if (fSeed1 < 0)
//    fSeed1 += 2147483563;
//  k = (int) (fSeed2 / 52774);
//  fSeed2 = 40692 * (fSeed2 - k * 52774) - k * 3791;
//  if (fSeed2 < 0)
//    fSeed2 += 2147483399;
//  iz = fSeed1 - fSeed2;
//  if (iz <= 0)
//    iz += 2147483562;
//  r = iz * 4.6566128e-10;
//  return r;
//}

//// returns a uniform deviate on the interval ]x1, x2].
//double ranuni (double x1, double x2)
//{
//  double ans = rndm ();
//  return x1 + (x2 - x1) * ans;
//}

//double rangaus (double sigma)
//{
////      Return a number distributed following a gaussian with sigma

//  double x, y, z, result;

//  y = rndm ();
//  z = rndm ();
//  x = z * 6.28318530717958623;
//  result = sigma * sin (x) * sqrt (-2 * log (y));
//  return result;
//}
