//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef DSPLIB_H
#define DSPLIB_H

#include <cmath>

#ifndef sqr
# define sqr(x) ((double)(x)*(double)(x))
#endif

#ifndef hypot2
# define hypot2(x,y) ((double)(x)*(double)(x)+(double)(y)*(double)(y))
#endif

#ifdef M_PI
# define pi    M_PI
#else
# define pi    (double)3.14159265358979312e+00
#endif

//#define e     (double)2.71828182845904509e+00
//#define pi    (double)3.14159265358979312e+00
//#define twopi (double)6.28318530717958623e+00

// Gaussian function FWHM constant (sigma to width conversion)
// sqrt(2*log(2))
#define GausFWHM 2.35482004503094933e+00

#define BITS2MAXINT(bits)  ((1<<((bits)-1)) - 1)
#define BITS2MININT(bits)  (-(1<<((bits)-1)))
#define BITS2MAXUINT(bits) ((1<<(bits)) - 1)
#define BITS2MINUINT(bits) 0

typedef enum {FIT_ALGO_ACCURATE=0, FIT_ALGO_COARSE=1} fit_algo_t;

// Find a,b for array of (x,y) of length n
//y = ax + b
void LinearFit (const double *x, const double *y, int n, double *a, double *b, double *c);
void LinearFitInt (const int *y, int n, double *a, double *b, double *c);

// Convolution
// size of output is l1+l2
// Integer+Integer -> Integer
void convolve (const int *x1, int l1, const int *x2, int l2, int *out);
// Integer+ShortInteger -> Integer
void convolve_i_s_i (const int *x1, int l1, const short int *x2, int l2, int *out);
// Double+Double -> Double
void convolve_d (const double *x1, int l1, const double *x2, int l2, double *out);
// Integer+Double -> Double
void convolve_id (const int *x1, int l1, const double *x2, int l2, double *out);

// 2nd derivative
void deriv2 (const int *x, int len, int *out);
/*
// windowing functions
// // i = 0, 1, ..., m
  double wf_barthann (double i, int m);
  double wf_barlett (double i, int m);
  double wf_blackman (double i, int m);
  double wf_blackmanharris (double i, int m);
  double wf_bohman (double i, int m);
  double wf_flattop (double i, int m);
  double wf_gauss (double i, int m, double r);
  double wf_gauss_075 (double i, int m);
  double wf_gauss_050 (double i, int m);
  double wf_gauss_025 (double i, int m);
  double wf_hamming (double i, int m);
  double wf_hann (double i, int m);
  double wf_nuttall (double i, int m);
  double wf_rect (double i, int m);
  double wf_triang (double i, int m);
  double wf_tukey (double i, int m, double r);
  double wf_tukey_075 (double i, int m);
  double wf_tukey_050 (double i, int m);
  double wf_tukey_025 (double i, int m);
*/
typedef enum {
    WF_BARTHANN,
    WF_BARLETT,
    WF_BLACKMAN,
    WF_BLACKMANHARRIS,
    WF_BOHMAN,
    WF_FLATTOP,
    WF_GAUSS_075,
    WF_GAUSS_050,
    WF_GAUSS_025,
    WF_HAMMING,
    WF_HANN,
    WF_NUTTALL,
    WF_RECT,
    WF_TRIANG,
    WF_TUKEY_075,
    WF_TUKEY_050,
    WF_TUKEY_025,
} fwinf_t;

double windowingFunction (fwinf_t win_type, double i, int m);

// Return noise floor in dB
double fft_noisefloor (double rmsnoise, int rank);

// Discrete Fourier Transform, real form, order N
void DFT (const double *x, int n, double *S, double *C);

//Inverse Discrete Fourier Transform, real form, order N
void IDFT (double *x, int n, const double *S, const double *C);

// Calculate power from rectangular FFT
void RECT2POW (const double *S, const double *C, int n, double *power);

// Calculate phase from rectangular FFT
void RECT2PHA (const double *S, const double *C, int n, double *phase);

// Complex Divide operation in rectangular form: (e+if)=(a+ib)/(c+id)
void Compdiv (double a, double b, double c, double d, double *e, double *f);

// Normalize to integral=1
void Normalize (double *x, int n);

// Scale to fit in <bits>-bit integer
void Make_Discrete (const double *x, int *out, int n, int bits);

// Arithmetic sum of array elements (int)
int intsum (const int *x, int len);

// Max/Min value of array elements (int)
double maxdbl (const double *x, int len);
double mindbl (const double *x, int len);
// DC value
double dcdbl (const double *x, int len);
// RMS value
double rmsdbl (const double *x, int len, double dc);

// Decibels to value conversion
double db2val (double db);

// Value to decibels conversion
double val2db (double val);

// An unnormalized gaussian function with mean=0 and sigma
// Divide result by sqrt(2*Pi) = 2.50662827463100024 if normalized output needed
double gauss (double x, double sigma);

////  Machine independent random number generator.
////  Produces uniformly-distributed floating points in [0,1]
////  Identical sequence on all machines of >= 32 bits.
////  Periodicity > 10**14
//  double rndm ();

//// returns a uniform deviate on the interval ]x1, x2].
//  double ranuni (double x1, double x2);

////      Return a number distributed following a gaussian with sigma
//  double rangaus (double sigma);

// Simulate detector pulse
// amplitude a, time t, risetime rt, falltime ft
double detpulse (double a, double t, double rt, double ft);

#endif                          /* DSPLIB_H */
