//
//    Beam Position Monitor
//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <iostream>
#include <algorithm>
#include <stdexcept>

#define _USE_MATH_DEFINES
#include <cmath>

#include <fftw.h>
#include <rfftw.h>

#include "fftengine.h"
#include "dsplib.h"
#include "TimeDomainData.h"

using namespace std;

void FFTEngine::update_wind_norm() {
    const fwinf_t wf = WF_HAMMING;
    //  const fwinf_t wf = WF_HANN;
    //  const fwinf_t wf = WF_BLACKMAN;
    //  const fwinf_t wf = WF_BLACKMANHARRIS;
    //  const fwinf_t wf = WF_GAUSS_025;
    wind_norm = 0;
    for(size_t i=0; i<N; i++)
        wind_norm += windowingFunction(wf, i, N) - (windowingFunction(wf, 0, N)+windowingFunction(wf, N-1, N))/2;
    wind_norm /= N;
    //  cerr << "Windowing norm: " << wind_norm << endl;
    //  cerr << "Window edges: " << windowingFunction(wf, 0, N) << ", " << windowingFunction(wf, N-1, N) << endl;
}

void RFFTPlan::resize(size_t n)
{
    if (n == N)
        return;
    delete fftwr_p;
    delete fftwr_inv_p;
    delete[] in;
    delete[] out;
    N = n;
//    cerr << "RFFTPlan("<<N<<")"<<endl;
    fftwr_p = rfftw_create_plan(N, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
    fftwr_inv_p = rfftw_create_plan(N, FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE);
    in = new fftw_real[N];
    out = new fftw_real[N];
}

//--------------------------------------------------------------
void CFFTPlan::resize(size_t n)
{
    if (n == N)
        return;
    delete fftw_p;
    delete fftw_inv_p;
    delete[] in;
    delete[] out;
    N = n;
//    cerr << "CFFTPlan("<<N<<")"<<endl;
    fftw_p = fftw_create_plan(N, FFTW_FORWARD, FFTW_ESTIMATE);
    fftw_inv_p = fftw_create_plan(N, FFTW_BACKWARD, FFTW_ESTIMATE);
    in = new fftw_complex[N];
    out = new fftw_complex[N];
}

//--------------------------------------------------------------

// Sample rate converter
// Interpolation algorythms:
// 0 - nearest neighbour
// 1 - linear
// 2 - quadratic
// 3 - cubic
// 4 - cosine
std::vector<double> RFFTEngine::resample(const std::vector<double> &d, double ratio, int algo, double oversample_factor)
{
    size_t interp_sz = (int)(3./ratio); // number of points reserved for cubic interpolator

    if (ratio <= 0 || ratio == 1 || d.size()<=interp_sz) return std::vector<double>(d);

    std::vector<double> v(d);
    v.resize((v.size()+1023) / 1024 * 1024, v[v.size()-1]); // align size
    if (oversample_factor!=1)
        v = oversample(v, oversample_factor);

    std::vector<double> overv(v);
    size_t sz = overv.size();
    size_t rsize = (int)((overv.size()-interp_sz) * ratio);
    v.resize(rsize);
    //  std::vector<double> result(rsize, 0);
    //  cerr << "rsize=" << rsize << endl;
    size_t i=0;
    while(i+interp_sz < rsize) {
        size_t ind1 = (int)(i/ratio);
        size_t ind2 = (int)((i+1)/ratio);
        size_t ind3 = (int)((i+2)/ratio);
        size_t ind4 = (int)((i+3)/ratio);
        if (ind4 >= sz) throw std::runtime_error("hehehe");
        double v0 = overv[ind1];
        double v1 = overv[ind2];
        double v2 = overv[ind3];
        double v3 = overv[ind4];
        double mu = 1.*i/ratio - ind1;
        if (algo == 0)
            v[i] = v0;
        else if (algo == 1) {
            v[i] = v0*(1-mu) + v1*mu; // linear
        }
        else if (algo == 2) {
            v[i] = v0; // TODO
        }
        else if (algo == 3) {
            double mu2 = mu*mu;
            // cubic interp
            double a0 = v3 - v2 - v0 + v1;
            double a1 = v0 - v1 - a0;
            double a2 = v2 - v0;
            double a3 = v1;
            v[i] = a0*mu*mu2+a1*mu2+a2*mu+a3;
        }
        else if (algo == 4) {
            // cosine interp
            double mu2 = (1-cos(mu*M_PI))/2;
            v[i] = v0*(1-mu2)+v1*mu2; // cosine
        } else v[i] = v0;
        //	cerr << "resampled["<<i<<"], ind=" << ind << endl;
        i++;
    }
    for(size_t k=i; k<v.size(); k++)
        v[k] = v[i];
    if (oversample_factor!=1)
        v = oversample(v, 1./oversample_factor);
    return v;
}

std::vector<double> RFFTEngine::oversample(const std::vector<double> &d, double ratio)
{
    //  if (d.size() != N)
    //	throw exc("Incompatible input data size for FFT");

    const size_t N = d.size();
    if (N==0 || ratio == 0) return std::vector<double>();

    //  cerr << "Allocating " << N << " order FFT" << endl;
    fftw_real *in = new fftw_real[N];
    fftw_real *out = new fftw_real[N];
    rfftw_plan fftwr_p = rfftw_create_plan(N, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);

    const size_t Nover = (int)(N*ratio);

    //  cerr << "Allocating " << Nover << " order reverse FFT" << endl;
    fftw_real *overin = new fftw_real[Nover];
    fftw_real *overout = new fftw_real[Nover];
    rfftw_plan fftwr_rev_p = rfftw_create_plan(Nover, FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE);

    //  cerr << "Computing forward FFT("<<N<<")" << endl;

    for(size_t i=0; i<N; i++)
        in[i] = d[i];
    rfftw_one(fftwr_p, in, out);

    //  cerr << "Original spectrum: "<< endl;
    //  for(size_t i=0; i<N; i++)
    //	cerr << "["<<i<<"]: "<< out[i] << endl;

    //  cerr << "Scaling spectrum: " << N << " to " << Nover << endl;

    for(size_t i=0; i<Nover; i++)
    {overin[i] = 0;overout[i] = 0;}
    overin[0] = out[0];  /* DC component */
    size_t max_N = (N+1)/2; if (max_N > (Nover+1)/2) max_N = (Nover+1)/2;
    //  cerr << "index: " << max_N << ", " << Nover - max_N << endl;
    for (size_t k = 1; k < max_N; ++k) { /* (k < N/2 rounded up) */
        overin[k] = out[k];
        overin[(Nover-k)] = out[N-k];
    }
    if ((N <= Nover) && (N % 2 == 0)) /* N is even */
        overin[N/2] = out[N/2];  /* Nyquist freq. */

    //  cerr << "Scaled spectrum: "<< endl;
    //  for(size_t i=0; i<Nover; i++)
    //	cerr << "["<<i<<"]: "<< overin[i] << endl;

    //  cerr << "Computing reverse FFT("<<Nover<<")" << endl;

    rfftw_one(fftwr_rev_p, overin, overout);

    std::vector<double> result(Nover, 0.0);
    for(size_t i=0; i<Nover; i++) {
        result[i] = overout[i] / N;//(sqrt(N) * sqrt(Nover));
    }

    //  cerr << "Oversampled data["<< Nover<<"]: "<< endl;
    //  for(size_t i=0; i<Nover; i++)
    //	cerr << "["<<i<<"]: "<< overout[i] << endl;

    //  cerr << "Deallocating" << endl;
    delete[] in;
    delete[] out;
    fftw_destroy_plan(fftwr_p);
    delete[] overin;
    delete[] overout;
    fftw_destroy_plan(fftwr_rev_p);

    return result;
}

// perform forward FFT for real data, result in halfcomplex format
std::vector<double> RFFTEngine::fft(const std::vector<double> &d) const
{
    if (d.size() != N)
        throw std::runtime_error("Incompatible input data size for FFT");
    for(size_t i=0; i<N; i++)
        plan.in[i] = d[i]; // * windowingFunction(WF_HAMMING, i, N) / wind_norm;
    rfftw_one(plan.fftwr_p, plan.in, plan.out);
    std::vector<double> result(N);
    for(size_t i=0; i<N; i++)
        result[i] = plan.out[i];
    return result;
}

// perform forward FFT halfcomplex data, result is real data
std::vector<double> RFFTEngine::ifft(const std::vector<double> &d) const
{
    if (d.size() != N)
        throw std::runtime_error("Incompatible input data size for FFT");
    for(size_t i=0; i<N; i++)
        plan.in[i] = d[i]; // * windowingFunction(WF_HAMMING, i, N) / wind_norm;
    rfftw_one(plan.fftwr_inv_p, plan.in, plan.out);
    std::vector<double> result(N);
    for(size_t i=0; i<N; i++)
        result[i] = plan.out[i];
    return result;
}
/*
std::vector<double> RFFTEngine::powDBFS(const std::vector<double> &d)
{
  if (d.size() != N) {
        resize(d.size());
//	throw exc("Incompatible input data size for FFT");
  }
  const size_t nre = N/2+1;
  for(size_t i=0; i<N; i++)
        plan.in[i] = d[i] * windowingFunction(WF_HAMMING, i, N) / wind_norm;
  rfftw_one(plan.fftwr_p, plan.in, plan.out);
  std::vector<double> result(nre,0);
// out is in halfcomplex format:
// 0<=i<=N/2, out[i]=Re(Xi)
// 0<i<N/2,   out[N-i]=Im(Xi)
  const fftw_real *out = plan.out;;
  result[0] = out[0]*out[0];  // DC component
  for (size_t k = 1; k < (N+1)/2; ++k)  // (k < N/2 rounded up)
    result[k] = out[k]*out[k] + out[N-k]*out[N-k];
  if (N % 2 == 0) // N is even
    result[N/2] = out[N/2]*out[N/2];  // Nyquist freq.

  for(size_t i=0; i<nre; i++) {
        result[i] = sqrt(result[i]) / N;
  }
  return result;
}
*/

FFTSpectrum RFFTEngine::getFFTSpectrum(const TimeDomainData &d)
{
    if (d.size() != N) {
        resize(d.size());
        //	throw exc("Incompatible input data size for FFT");
    }
    const size_t nre = N/2+1;
    for(size_t i=0; i<N; i++)
        plan.in[i] = d[i] * windowingFunction(WF_HAMMING, i, N) / wind_norm;
    rfftw_one(plan.fftwr_p, plan.in, plan.out);
    FFTSpectrum result;
    result.resize(nre);
    // out is in halfcomplex format:
    // 0<=i<=N/2, out[i]=Re(Xi)
    // 0<i<N/2,   out[N-i]=Im(Xi)
    const fftw_real *out = plan.out;;
    result[0].power = out[0]*out[0];  /* DC component */
    for (size_t k = 1; k < (N+1)/2; ++k) { /* (k < N/2 rounded up) */
        result[k].power = out[k]*out[k] + out[N-k]*out[N-k];
        result[k].phase = atan2(out[N-k], out[k]);
    }
    if (N % 2 == 0) { /* N is even */
        result[N/2].power = out[N/2]*out[N/2];  /* Nyquist freq. */
    }

    double samplingFreq = d.samplingFreq;
    for(size_t i=0; i<nre; i++) {
        result[i].power = sqrt(result[i].power) / N;
        result[i].freq = samplingFreq/N*i;
    }
    //  result.print();
    return result;
}

